"""Exports data to pandas DataFrame."""

from typing import Any

from pandas import DataFrame, json_normalize, merge

from ..providers.pff.schema.tracking import PFF_Frame
from ..visualization.view import pydantic_dump_options


def _build_df(obj_list: list, dump_options: dict[str, Any] | None = None) -> DataFrame:
    if dump_options is None:
        dump_options = {}
    options = {**pydantic_dump_options, **dump_options}
    data = [obj.model_dump(**options) for obj in obj_list]
    return json_normalize(data, sep='_')


def _extract_players_ball_from_frames(
    frames: list[PFF_Frame], *, smoothed: bool = False
) -> list[dict[str, Any]]:
    ball_key = 'ball_with_kalman' if smoothed else 'ball'
    home_key = 'home_players_with_kalman' if smoothed else 'home_players'
    away_key = 'away_players_with_kalman' if smoothed else 'away_players'

    player_dump_options = {'exclude': {'shirt_confidence', 'visibility'}}
    ball_dump_options = {'exclude': {'visibility'}}

    extracted_frames = []

    for frame in frames:
        home_players = getattr(frame, home_key, [])
        away_players = getattr(frame, away_key, [])

        home = [
            {**player.model_dump(**player_dump_options), 'team': 'home'}
            for player in home_players
        ]
        away = [
            {**player.model_dump(**player_dump_options), 'team': 'away'}
            for player in away_players
        ]

        ball = getattr(frame, ball_key, None)
        if ball:
            if isinstance(ball, list):
                ball = ball[0]
            ball_data = [ball.model_dump(**ball_dump_options)]
        else:
            ball_data = []

        players = home + away
        id = frame.frame_id

        extracted_frames.append({'frame_id': id, 'players': players, 'ball': ball_data})

    return extracted_frames


def _build_players_ball_df(
    frames: list[PFF_Frame], *, smoothed: bool = False
) -> DataFrame:
    dump_options = {'include': {'match_id', 'frame_id', 'period', 'elapsed_seconds'}}

    frame_df = _build_df(frames, dump_options)

    coordinates = _extract_players_ball_from_frames(frames, smoothed=smoothed)
    players_df = json_normalize(
        data=coordinates,
        record_path='players',
        meta=['frame_id'],
        sep='_',
    )
    ball_df = json_normalize(
        data=coordinates,
        record_path='ball',
        record_prefix='ball_',
        meta=['frame_id'],
        sep='_',
    )

    players_ball_df = merge(players_df, ball_df, on='frame_id')
    return frame_df.merge(players_ball_df, on='frame_id')


def _build_metadata_df(frames: list[PFF_Frame]) -> DataFrame:
    dump_options = {
        'exclude': {
            'ball',
            'ball_with_kalman',
            'home_players',
            'away_players',
            'home_players_with_kalman',
            'away_players_with_kalman',
        },
        'exclude_unset': False,
        'exclude_defaults': False,
        'exclude_none': False,
    }

    df = _build_df(frames, dump_options)

    # Define the final columns
    final_columns = [
        'match_id',
        'frame_id',
        'period',
        'elapsed_seconds',
        'home_ball',
        'event_id',
        'event_game_event_type',
        'event_setpiece_type',
        'event_player_id',
        'event_team_id',
        'event_start_frame',
        'event_end_frame',
        'possession_id',
        'possession_possession_event_type',
        'possession_start_frame',
        'possession_end_frame',
        'sequence',
        'version',
        'video_time_milli',
    ]

    df = df.loc[:, final_columns]

    # Rename columns
    columns_to_rename = {
        'home_ball': 'home_has_possession',
        'event_game_event_type': 'event_type',
        'possession_possession_event_type': 'possession_type',
    }
    df = df.rename(columns=columns_to_rename)

    return df


def pff_frames_to_dataframe(
    frames: list[PFF_Frame], **kwargs
) -> tuple[DataFrame, DataFrame]:
    """Splits PFF frames into metadata and player+ball dfs.

    The PFF Frames collection is transformed into a DataFrame that contains
    frame metadata and another DataFrame with players and ball coordinates.

    Parameters:
    - frames: List of PFF_Frame objects.

    Returns:
    - A tuple containing the metadata DataFrame and the players plus ball DataFrame.
    """
    metadata_df = _build_metadata_df(frames)
    players_ball_df = _build_players_ball_df(frames)

    return metadata_df, players_ball_df



from pydantic import BaseModel, ValidationError, field_validator, model_validator

PFF_API_VERSION = '0.6.100'
PFF_FRAME_VERSION = '4.0.0'  # TODO: Get version num


class PFFResponse(BaseModel):
    _version: str


class PFFEventResponse(PFFResponse):
    _event_version: str = PFF_API_VERSION

    @model_validator(mode='before')
    def set_data_version(cls, values: dict):
        event_version = values.get('_event_version')
        if not event_version or not isinstance(event_version, str):
            raise ValidationError('`event version` tampered with')
        # TODO: Add format verification
        values['_version'] = event_version
        return values


class PFFFrameResponse(PFFResponse):
    _frame_version: str = PFF_FRAME_VERSION

    @model_validator(mode='before')
    def set_data_version(cls, values: dict):
        frame_version = values.get('_frame_version')
        if not frame_version or not isinstance(frame_version, str):
            raise ValidationError('`frame version` tampered with')
        # TODO: ADd format verification
        values['_version'] = frame_version
        return values


class AvailableCompetitions(PFFEventResponse):
    id: int
    name: str
    seasons: list[str]

    @field_validator('seasons', mode='before')
    def populate_seasons(cls, v):
        return [d['season'] for d in v]


class Competition(PFFEventResponse): ...


class Competitions(PFFEventResponse): ...



import bz2
import pickle
from pathlib import Path
from typing import Callable, Generator

import orjson
import pandas as pd
import structlog
from tqdm.auto import tqdm

logger = structlog.get_logger()


def check_path(path: str | Path):
    """Ensures that the path exists.

    :param data_dir: The path to the data directory.
    :raises FileNotFoundError: If the data directory does not exist.
    """
    if not Path(path).exists():
        raise FileNotFoundError(f'Path {path} does not exist')


def read_json(file_path: str | Path) -> list[dict]:
    """Loads a JSON file from `file_path`.

    :param file_path: The path to the JSON file.
    :return: The file contents
    """
    try:
        path = Path(file_path)
        with path.open('rb') as file:
            return orjson.loads(file.read())
    except Exception as exc:
        raise IOError(f'Failed to parse data from {file_path}') from exc


def read_jsonl(
    file_path: str | Path, *, record_filter: Callable | None
) -> Generator[list[dict], None, None]:
    """Loads a jsonl file and returns its contents as a list of dictionaries.

    :param file_path: The path to the .jsonl file.
    :param filter: A function to filter the records as they are read.
    :return: A list of dictionaries, where each dictionary corresponds to a JSON object
    from an entire line in the file.
    :raises FileNotFoundError: If the file does not exist.
    :raises IOError: If the file cannot be read or parsed.
    """
    try:
        path = Path(file_path)
        with path.open('rb') as file:
            for line in file:
                record = orjson.loads(line)
                if record_filter and record_filter(record) is False:
                    continue
                yield record
    except Exception as exc:
        raise IOError(f'Failed to parse data from {file_path}') from exc


def read_bz2(
    file_path: str | Path, encoding: str = 'utf-8', *, record_filter: Callable | None
) -> Generator[list[dict], None, None]:
    """Loads and decompresses a .bz2 file. The expected contents are jsonl files with
    tracking data from PFF.

    :param file_path: The path to the .bz2 file.
    :param encoding: The encoding to use when decoding the bytes. Defaults to 'utf-8'.
    :return: The decompressed content as bytes if decode is False, otherwise as string.
    :raises IOError: If the file cannot be read or decompressed.
    """
    try:
        path = Path(file_path)
        with bz2.open(path, 'rb') as file:
            file_contents = file.read()
        decoded_jsonl = file_contents.decode(encoding)
        lines = filter(None, decoded_jsonl.strip().split('\n'))
        for line in lines:
            record = orjson.loads(line)
            if record_filter and record_filter(record) is False:
                continue
            yield record
    except Exception as exc:
        raise IOError(f'Failed to read or decompress {file_path}') from exc


def read_csv(file_path: str | Path) -> pd.DataFrame:
    """Loads a CSV file from `file_path`.

    :param file_path: The path to the CSV file.
    :return: The file contents
    """
    try:
        path = Path(file_path)
        return pd.read_csv(path)
    except Exception as exc:
        raise IOError('Failed to parse data from {file_path}') from exc


def read_pickle(file_path: str | Path) -> pd.DataFrame:
    """Loads a pickle file from `file_path`.

    :param file_path: The path to the pickle file.
    :return: The file contents
    """
    try:
        path = Path(file_path)
        with path.open('rb') as file:
            return pickle.load(file)
    except Exception as exc:
        raise IOError('Failed to parse data from {file_path}') from exc


def build_path(data_dir: str, competition_name: str | None, season: str | None) -> Path:
    """Builds the path to the data file for a given match.
    :param data_dir: The base directory where the data is stored.
    :param competition_name: The name of the competition.
    :param season: The season of the competition.
    :param match_id: The ID of the match.
    :return: The path to the data file.
    """
    if (not competition_name and season) or (competition_name and not season):
        print(
            f'Both competition_name and season must be provided; fallback to {data_dir}'
        )

    path = Path(data_dir)
    if competition_name and season:
        path = path / competition_name / season

    check_path(path)
    return path


def get_filetype(directory: Path, match_id: int) -> str:
    """Searches for the file type for match `match_id`, either .jsonl or .bz2.

    :param directory: The directory to search in.
    :param match_id: The match ID to search for in the filename.
    :return: The filename if found, otherwise raises a FileNotFoundError.
    :raises FileNotFoundError: If no file with the specified match_id and suffix exists.
    """
    for file in Path(directory).iterdir():
        if (
            file.is_file()
            and (file.stem == str(match_id) or file.stem == str(match_id) + '.jsonl')
            and file.suffix in ['.jsonl', '.bz2']
        ):
            return file.name

    raise FileNotFoundError(
        f'No file found for match {match_id} with .jsonl or .bz2 in {directory}'
    )


def get_path(
    data_dir: str,
    match_id: int,
    competition_name: str | None = None,
    season: str | None = None,
) -> Path:
    """Combines building the directory path and checking if the file exists.

    :param data_dir: The base directory where the data is stored.
    :param match_id: The match ID to search for.
    :param competition_name: The name of the competition.
    :param season: The season of the competition.
    :return: A Path object representing the file.
    :raises FileNotFoundError: If the directory or file does not exist.
    """
    dirname = build_path(data_dir, competition_name, season)
    filename = get_filetype(dirname, match_id)
    return dirname / filename


def find_available_matches(
    data_dir: str, competition_name: str | None = None, season: str | None = None
) -> list[str]:
    """Returns a list of available matches for `competition_name` for a given season.

    :param data_dir: The base directory where the data is stored.
    :param competition_name: The name of the competition.
    :param season: The season of the competition.
    """
    path = build_path(data_dir, competition_name, season)

    return [
        file.name
        for file in path.iterdir()
        if file.is_file() and file.suffix in ['.jsonl', '.bz2']
    ]


def get_frames(
    data_dir: str,
    match_id: int,
    *,
    competition_name: str | None = None,
    season: str | None = None,
    record_filter: Callable | None = None,
) -> list[dict]:
    """Local client function to retrieve match data for a given match_id.

    :param data_dir: The base directory where the data is stored.
    :param competition_name: The name of the competition (optional).
    :param season: The season of the competition (optional).
    :param match_id: The ID of the match to retrieve data for.
    :return: A list of dictionaries containing the frames from PFF.
    """
    file_path = get_path(data_dir, match_id, competition_name, season)

    records = []

    if file_path.suffix == '.jsonl':
        file_data = read_jsonl(file_path, record_filter=record_filter)
    elif file_path.suffix == '.bz2':
        file_data = read_bz2(file_path, record_filter=record_filter)
    else:
        raise Exception(f'File type {file_path.suffix} not supported')

    for line in tqdm(
        file_data,
        unit=' frames',
        desc=f'Loading frames from match {match_id}',
        leave=False,
    ):
        records.append(line)

    return records



"""Schema for Atomic SPADL."""

from typing import Annotated

from pydantic import BaseModel, ConfigDict, Field

from .types import SpadlAction, SpadlBodyPart


class AtomicSpadl(BaseModel):
    match_id: int
    event_id: int
    period_id: int
    player_id: int
    team_id: int
    time_seconds: Annotated[float, Field(strict=True, gt=0)]
    x: Annotated[float, Field(strict=True, ge=0, le=105)]
    y: Annotated[float, Field(strict=True, ge=0, le=68)]
    dx: Annotated[float, Field(strict=True, ge=-105, le=105)]
    dy: Annotated[float, Field(strict=True, ge=-68, le=68)]
    action_type: SpadlAction
    bodypart: SpadlBodyPart

    model_config = ConfigDict(use_enum_values=True)

    def __eq__(self, other):
        return self.match_id == other.match_id and self.event_id == other.event_id


#     def __repr__(self):
#         return (
#             f'{self.action_type.name}'
#             f' by player {self.player_id} ({self.team_id})'
#             f' at {self.time_seconds}s'
#         )



from functools import lru_cache

from .query_utils import build_full_query

game = """
query game($id: ID!) {
    game(id: $id) {
        ...GameFragment
        gameEvents {
            ...GameEventFragment
        }
    }
}
"""

game_events = """
query gameEvents(
    $gameId: ID!,
    $otherPlayerId: ID,
    $playerId: ID,
    $playerOffId: ID,
    $playerOnId: ID,
    $pressurePlayerId: ID,
    $sortField: GameEventSortFields,
    $sortOrder: SortOrder
) {
    gameEvents(
        gameId: $gameId,
        otherPlayerId: $otherPlayerId,
        playerId: $playerId,
        playerOffId: $playerOffId,
        playerOnId: $playerOnId,
        pressurePlayerId: $pressurePlayerId,
        sortField: $sortField,
        sortOrder: $sortOrder
    ) {
        ...GameEventFragment
    }
}
"""

game_event = """
query gameEvent($id: ID!) {
    gameEvent(id: $id) {
       ...GameEventFragment 
    }
}
"""

pass_event = """
query passingEvent($id: ID!) {
    passingEvent(id: $id) {
        ...PassingEventFragment
    }
}
"""

available_competitions = """
query competitions(
    $sortField: CompetitionSortFields,
    $sortOrder: SortOrder
) {
    competitions(
        sortField: $sortField,
        sortOrder: $sortOrder
    ) {
        id
        name
        seasons: availableSeasons {
            ...CompetitionSeasonFragment
        }
    }
}
"""

competitions = """
query competitions(
  $sortField: CompetitionSortFields,
  $sortOrder: SortOrder
) {
  competitions(
    sortField: $sortField,
    sortOrder: $sortOrder
  ) {
    availableSeasons {
      ...CompetitionSeasonFragment
    }
    id
    insertedAt
    games {
      ...CompetitionGameFragment
    }
    lastSeason {
      ...CompetitionSeasonFragment
    }
    logoIcon
    name
    teams {
      ...TeamFragment
    }
    updatedAt
  }
}
"""

competition = """
query competition($id: ID!) {
  competition(id: $id) {
    availableSeasons {
      ...CompetitionSeasonFragment
    }
    id
    insertedAt
    games {
      ...CompetitionGameFragment
    }
    lastSeason {
      ...CompetitionSeasonFragment
    }
    logoIcon
    name
    teams {
      ...TeamFragment
    }
    updatedAt
  }
}
"""

players_by_competition = """
query competition ($id: ID!) {
  competition (id: $id) {
    games {
      rosters {
        player {
          ...PlayerFullFragment
        }
        shirtNumber
        game {
          id
        }
      }
    }
  } 
}
"""

player = """
query player($id: ID!) {
  player(id: $id) {
    ...PlayerFullFragment
    rosters {
      ...RosterShortFragment
    }
  }
}
"""

player_dribble_metrics = """
query playerDribbleMetricsReport(
  $competitionId: ID!,
  $playerId: Int!,
  $season: String!
) {
    playerDribbleMetricsReport(
        competitionId: $competitionId,
        playerId: $playerId,
        season: $season
    ) {
        ...DribbleMetricsFragment
    }
}
"""

account_active = """
query accountActive {
  accountActive
}
"""

version = """
query version {
  version
}
"""


@lru_cache()
def get_queries() -> dict[str, str]:
    queries = {
        'game': game,
        'game_events': game_events,
        'game_event': game_event,
        'pass_event': pass_event,
        'competitions': competitions,
        'competition': competition,
        'players_by_competition': players_by_competition,
        'player': player,
        'account_active': account_active,
        'version': version,
        'player_dribble_metrics': player_dribble_metrics,
        'available_competitions': available_competitions,
    }

    for query_name, query in queries.items():
        queries[query_name] = build_full_query(query)

    return queries


queries = get_queries()



import re

from .fragments import fragments


def extract_fragment_names(query_or_fragment: str) -> set[str]:
    """Extracts fragment names from a given GraphQL query or fragment.

    :param query_or_fragment: The GraphQL query or fragment string.
    :type query_or_fragment: str
    :return: A set of fragment names used in the query or fragment.
    :rtype: set[str]
    """
    pattern = re.compile(r'\.\.\.(\w+)')  # extracts WORD from  ...WORD
    return set(pattern.findall(query_or_fragment))


def resolve_fragments(
    fragments_to_resolve: set[str], resolved_fragments: set[str] | None = None
) -> set[str]:
    """Recursively resolve all nested fragments required for a given fragment.

    :param fragments_to_resolve: The set of fragment names that need to be resolved.
    :param resolved_fragments: A set of resolved fragment names (used in recursion).
    :return: A set of all required fragment names.
    :raises ValueError: If a fragment name is not found in the fragments dictionary.
    """
    if resolved_fragments is None:
        resolved_fragments = set()

    for fragment_name in fragments_to_resolve:
        if fragment_name in resolved_fragments:
            continue

        fragment_body = fragments[fragment_name]
        if not fragment_body:
            raise ValueError(f"Fragment '{fragment_name}' not found.")

        resolved_fragments.add(fragment_name)
        nested_fragments = extract_fragment_names(fragment_body)
        resolve_fragments(nested_fragments, resolved_fragments)

    return resolved_fragments


def clean_graphql_query(query: str) -> str:
    """Cleans a GraphQL `query` by removing comments, extra whitespaces, and spacing.

    :param query: The GraphQL query string.
    :type query: str
    :return: The cleaned GraphQL query string.
    :rtype: str
    """
    query = re.sub(r'#.*', '', query)
    query = re.sub(r'\s+', ' ', query)
    query = re.sub(r'\s*([{}()\[\],:])\s*', r'\1', query)
    return query.strip()


def build_full_query(query: str) -> str:
    """Builds the full GraphQL query by resolving and appending all necessary fragments.

    :param query: The base GraphQL query string.
    :type query: str
    :return: The full GraphQL query string with all required fragments.
    :rtype: str
    """
    fragments_used = extract_fragment_names(query)
    fragments_required = resolve_fragments(fragments_used)

    fragment_section = '\n'.join(
        [fragments[fragment_name] for fragment_name in fragments_required]
    )

    full_query = f'{fragment_section}\n{query}'
    return clean_graphql_query(full_query)



"""Calculates speed and acceleration vectors on PFF tracks."""

from typing import cast

from pandas import DataFrame
from tqdm import tqdm


def _compute_derivatives(df: DataFrame) -> DataFrame:
    df = df.sort_values('frame_id')

    df['time_diff'] = df['elapsed_seconds'].diff()

    df['vx'] = df['x'].diff() / df['time_diff']
    df['vy'] = df['y'].diff() / df['time_diff']
    df['ax'] = df['vx'].diff() / df['time_diff']
    df['ay'] = df['vy'].diff() / df['time_diff']
    df['speed'] = (df['vx'] ** 2 + df['vy'] ** 2) ** 0.5

    return df.drop(columns=['time_diff'])


def _compute_ball_derivatives(df: DataFrame) -> DataFrame:
    ball_df = df.drop_duplicates(subset='frame_id', keep='first')

    ball_df = ball_df[['frame_id', 'elapsed_seconds', 'ball_x', 'ball_y', 'ball_z']]
    ball_df = cast(DataFrame, ball_df)

    ball_df = ball_df.sort_values('elapsed_seconds')

    ball_df['time_diff'] = ball_df['elapsed_seconds'].diff()

    ball_df['ball_vx'] = ball_df['ball_x'].diff() / ball_df['time_diff']
    ball_df['ball_vy'] = ball_df['ball_y'].diff() / ball_df['time_diff']
    ball_df['ball_vz'] = ball_df['ball_z'].diff() / ball_df['time_diff']
    ball_df['ball_ax'] = ball_df['ball_vx'].diff() / ball_df['time_diff']
    ball_df['ball_ay'] = ball_df['ball_vy'].diff() / ball_df['time_diff']
    ball_df['ball_az'] = ball_df['ball_vz'].diff() / ball_df['time_diff']
    ball_df['ball_speed'] = (
        ball_df['ball_vx'] ** 2 + ball_df['ball_vy'] ** 2 + ball_df['ball_vz'] ** 2
    ) ** 0.5

    ball_df = ball_df[
        [
            'frame_id',
            'ball_vx',
            'ball_vy',
            'ball_vz',
            'ball_speed',
            'ball_ax',
            'ball_ay',
            'ball_az',
        ]
    ]

    return df.merge(ball_df, on='frame_id', how='left')


def add_players_speed(frames_df: DataFrame) -> DataFrame:
    tqdm.pandas(desc='Adding player speed, velocity, and acceleration in each frame')
    return (
        frames_df.groupby(['period', 'team', 'shirt'], as_index=False)
        .progress_apply(_compute_derivatives)
        .reset_index(drop=True)
    )


def add_ball_speed(frames_df: DataFrame) -> DataFrame:
    tqdm.pandas(desc='Adding ball speed, velocity, and acceleration in each frame')
    return (
        frames_df.groupby('period', as_index=False)
        .progress_apply(_compute_ball_derivatives)
        .reset_index(drop=True)
    )



from .api_client import execute_query, get_client
from .queries import queries


def account_active(
    pff_api_url: str | None = None, pff_api_key: str | None = None
) -> dict:
    query = queries['account_active']
    client = get_client(pff_api_url, pff_api_key)
    return execute_query(query, client=client)


def version(pff_api_url: str | None = None, pff_api_key: str | None = None) -> dict:
    query = queries['version']
    client = get_client(pff_api_url, pff_api_key)
    return execute_query(query, client=client)



from . import logging as logging
from .api.auth import api_version as api_version
from .api.auth import validate_api_key as validate_api_key
from .api.competition import get_available_competitions as get_available_competitions
from .api.competition import get_competition as get_competition
from .api.competition import get_competitions as get_competitions
from .api.match import get_enhanced as get_enhanced
from .api.match import get_frames as get_frames
from .api.match import get_match as get_match
from .api.match import get_match_event as get_match_event
from .api.match import get_match_events as get_match_events
from .export import export as export  # LOL
from .providers.pff.local import loader as loader
from .providers.pff.schema.event import PFF_Game as PFF_Game
from .providers.pff.schema.tracking import Frame_Event as Frame_Event
from .visualization.video import video as video
from .visualization.view import view as view

# def _repr_html_() -> str:
#     # TODO: Switch to jinja template
#     return ''

# If there is a connection, try to fetch api_version.
# Match it with provder.response_type.PFF_EVENT_VERSION



from enum import Enum


class SpadlAction(Enum):
    PASS = 'pass'
    CROSS = 'cross'
    THROW_IN = 'throw_in'
    FREEKICK_CROSSED = 'freekick_crossed'
    FREEKICK_SHORT = 'freekick_short'
    CORNER_CROSSED = 'corner_crossed'
    CORNER_SHORT = 'corner_short'
    TAKE_ON = 'take_on'
    FOUL = 'foul'
    TACKLE = 'tackle'
    INTERCEPTION = 'interception'
    SHOT = 'shot'
    SHOT_PENALTY = 'shot_penalty'
    SHOT_FREEKICK = 'shot_freekick'
    KEEPER_SAVE = 'keeper_save'
    KEEPER_CLAIM = 'keeper_claim'
    KEEPER_PUNCH = 'keeper_punch'
    KEEPER_PICK_UP = 'keeper_pick_up'
    CLEARANCE = 'clearance'
    BAD_TOUCH = 'bad_touch'
    NON_ACTION = 'non_action'
    DRIBBLE = 'dribble'
    GOALKICK = 'goalkick'
    RECEIVAL = 'receival'
    OUT = 'out'
    OFFSIDE = 'offside'
    GOAL = 'goal'
    OWN_GOAL = 'owngoal'
    YELLOW_CARD = 'yellow_card'
    RED_CARD = 'red_card'
    CORNER = 'corner'
    FREEKICK = 'freekick'


class SpadlBodyPart(Enum):
    HEAD = 'head'
    FOOT = 'foot'
    OTHER = 'other'
    HEAD_OTHER = 'head_other'
    LEFT_FOOT = 'foot_left'
    RIGHT_FOOT = 'foot_right'



"""Functions to visualize gandula objects."""

from typing import overload

from ..providers.pff.schema.event import PFF_Event, PFF_PossessionEvent
from ..providers.pff.schema.tracking import PFF_Frame
from ..schemas.atomic_spadl.types import SpadlAction
from .frame import view_frame_sequence, view_pff_frame

pydantic_dump_options = {
    'exclude_none': True,
    'exclude_defaults': True,
    'exclude_unset': True,
}


@overload
def _view_pff_event(events: PFF_Event | PFF_PossessionEvent, **kwargs) -> dict: ...


@overload
def _view_pff_event(
    events: list[PFF_Event | PFF_PossessionEvent], **kwargs
) -> list[dict]: ...


def _view_pff_event(
    events: PFF_Event | PFF_PossessionEvent | list[PFF_Event | PFF_PossessionEvent],
    **kwargs,
) -> dict | list[dict]:
    options = {**pydantic_dump_options, **kwargs}

    if isinstance(events, list):
        return [evt.model_dump(**options) for evt in events]

    return events.model_dump(**options)


def _view_spadl_action(action: SpadlAction | list[SpadlAction], **kwargs) -> dict: ...


def view(objs, **kwargs):
    """Single entry point for visualizing gandula objects.

    Parameters
    ----------
    objs : PFF_Event | PFF_PossessionEvent | PFF_Frame | SpadlAction
        The object to be visualized.
    kwargs : dict
        Options to be passed to the view function.
    """
    obj = objs[0] if isinstance(objs, list) else objs

    if isinstance(obj, PFF_Event) or isinstance(obj, PFF_PossessionEvent):
        return _view_pff_event(objs, **kwargs)
    if isinstance(obj, PFF_Frame):
        if isinstance(objs, list):
            return view_frame_sequence(objs, **kwargs)
        return view_pff_frame(objs, **kwargs)
    if isinstance(obj, SpadlAction):
        return _view_spadl_action(objs, **kwargs)

    if hasattr(obj, 'model_dump'):
        options = {**pydantic_dump_options, **kwargs}
        if isinstance(objs, list):
            return [o.model_dump(**options) for o in objs]
        return obj.model_dump(**options)

    raise ValueError(f'Cannot view object of type {type(obj)}')



import logging
import sys

import orjson
import structlog

shared_log_processors = [
    structlog.contextvars.merge_contextvars,
    structlog.processors.add_log_level,
    structlog.processors.format_exc_info,
    structlog.processors.StackInfoRenderer(),
    structlog.processors.TimeStamper(fmt='iso', utc=True),
]


if not sys.stderr.isatty() and hasattr(sys.stderr, 'buffer'):
    processors = shared_log_processors + [
        structlog.processors.dict_tracebacks,
        structlog.processors.JSONRenderer(serializer=orjson.dumps),
    ]
    logger_factory = structlog.BytesLoggerFactory()
else:
    processors = shared_log_processors + [structlog.dev.ConsoleRenderer()]
    logger_factory = structlog.PrintLoggerFactory()

structlog.configure(
    processors=processors,
    logger_factory=logger_factory,
    wrapper_class=structlog.make_filtering_bound_logger(logging.INFO),
    context_class=dict,
    cache_logger_on_first_use=False,
)



import ast
from collections.abc import Callable
from pathlib import Path

import structlog
from tqdm.auto import tqdm

from ..providers.pff.api import match
from ..providers.pff.local import loader as local_loader
from ..providers.pff.schema.event import PFF_Event, PFF_Game, PFF_Roster
from ..providers.pff.schema.tracking import PFF_Frame
from ..utils.pff_enhancer import (
    enhance_match,
    filter_frames_with_actionable_events,
    get_enhanceable_event_ids,
)

logger = structlog.get_logger()


def get_match_events(
    match_id: int, *, api_url: str | None = None, api_key: str | None = None
) -> list[PFF_Event]:
    try:
        events = match.get_match_events(
            match_id, pff_api_url=api_url, pff_api_key=api_key
        )
        return [PFF_Event.model_validate(event) for event in events['gameEvents']]
    except Exception as exc:
        raise Exception(f'Error getting match events for match_id={match_id}') from exc


def get_match(
    match_id: int,
    *,
    api_url: str | None = None,
    api_key: str | None = None,
    events_path: str | None = None,
) -> PFF_Game:
    try:
        if events_path is None:
            logger.debug(f'Getting match {match_id} from API')
            pff_match = match.get_match(
                match_id, pff_api_url=api_url, pff_api_key=api_key
            )
            logger.debug(f'Validating match {match_id}')
            valid_match = PFF_Game.model_validate(pff_match['game'])
            return valid_match

        # TODO: Refactor this section
        logger.debug(f'Getting match {match_id} from local path {events_path}')
        pff_events = local_loader.read_json(Path(events_path) / 'events.json')
        pff_events = [
            PFF_Event.model_validate(event)
            for event in pff_events
            if event['gameId'] == int(match_id)
        ]

        pff_rosters = local_loader.read_csv(Path(events_path) / 'rosters.csv')
        pff_rosters.rename(columns={'game_id': 'gameId'}, inplace=True)
        pff_rosters = pff_rosters[pff_rosters['gameId'] == int(match_id)]
        pff_rosters['player'] = pff_rosters['player'].apply(ast.literal_eval)
        pff_rosters['team'] = pff_rosters['team'].apply(ast.literal_eval)
        pff_rosters = pff_rosters.to_dict(orient='records')
        pff_rosters = [PFF_Roster.model_validate(roster) for roster in pff_rosters]

        pff_metadata = local_loader.read_csv(Path(events_path) / 'metadata.csv')
        pff_metadata = pff_metadata[pff_metadata['id'] == int(match_id)]
        columns_to_convert = [
            'awayTeam',
            'awayTeamKit',
            'competition',
            'homeTeam',
            'homeTeamKit',
            'stadium',
            'videos',
        ]
        for col in columns_to_convert:
            pff_metadata[col] = pff_metadata[col].apply(ast.literal_eval)
        pff_metadata = pff_metadata.to_dict(orient='records')[0]

        pff_match = {'game': pff_metadata}
        pff_match['game']['gameEvents'] = pff_events
        pff_match['game']['allRosters'] = pff_rosters

        return PFF_Game.model_validate(pff_match['game'])
    except Exception as exc:
        raise Exception(f'Error getting match for match_id={match_id}') from exc


def get_match_event(
    event_id: int, *, api_url: str | None = None, api_key: str | None = None
) -> PFF_Event:
    try:
        event = match.get_match_event(
            event_id, pff_api_url=api_url, pff_api_key=api_key
        )
        return PFF_Event.model_validate(event)
    except Exception as exc:
        raise Exception(f'Error getting match event for event_id={event_id}') from exc


def get_frames(
    data_dir: str,
    match_id: int,
    *,
    competition_name: str | None = None,
    season: str | None = None,
    record_filter: Callable | None = None,
) -> list[PFF_Frame]:
    try:
        frames = local_loader.get_frames(
            data_dir,
            match_id,
            competition_name=competition_name,
            season=season,
            record_filter=record_filter,
        )

        frames = [
            PFF_Frame.model_validate(frame)
            for frame in tqdm(
                frames, desc='Validating frames', unit=' frames', leave=False
            )
        ]

        return frames
    except Exception as exc:
        logger.error(f'Error getting frames for match_id={match_id}', exc_info=exc)
        raise Exception(f'Error getting frames for match_id={match_id}') from exc


def get_enhanced(
    match_id: int,
    data_dir: str,
    *,
    competition_name: str | None = None,
    season: str | None = None,
    events_path: str | None = None,
    record_filter: Callable[[dict], dict | None]
    | None = filter_frames_with_actionable_events,
) -> tuple[PFF_Game, list[PFF_Frame]]:
    pff_match = get_match(match_id, events_path=events_path)
    event_ids, possession_event_ids = get_enhanceable_event_ids(pff_match.gameEvents)

    frames = get_frames(
        data_dir,
        match_id,
        competition_name=competition_name,
        season=season,
        record_filter=record_filter,
    )

    updated_match = enhance_match(pff_match, frames)

    return updated_match, frames



location = """
fragment LocationFragment on Location {
    eventModule
    id
    name
    x
    y
}
"""

player_id = """
fragment PlayerFragment on Player {
    id
}
"""

nation = """
fragment NationFragment on Nation {
    country
    federation {
        ...FederationFragment
    }
    fifaCode
    id
    iocCode
}
"""

federation = """
fragment FederationFragment on Federation {
    id
    abbreviation
    confederation {
        ...ConfederationFragment
    }
    country
    englishName
    name
}
"""

confederation = """
fragment ConfederationFragment on Confederation {
    id
    abbreviation
    name
}
"""

player_full = """
fragment PlayerFullFragment on Player {
    id
    firstName
    lastName
    nickname
    dob
    preferredFoot
    height
    weight
    gender
    positionGroupType
    nationality{
        ...NationFragment
    }
    secondNationality{
        ...NationFragment
    }
    countryOfBirth{
        ...NationFragment
    }
}
"""

roster_short = """
fragment RosterShortFragment on Roster {
    id
    game {
        id
    }
    positionGroupType
    shirtNumber
    started
    team {
        ...TeamFragment
    }
}
"""

roster = """
fragment RosterFragment on Roster {
    id
    game {
        id
    }
    player {
        ...PlayerFullFragment
    }
    positionGroupType
    shirtNumber
    started
    team {
        ...TeamFragment
    }
}
"""

team = """
fragment TeamFragment on Team {
    id
    name
    shortName
}
"""

kit = """
fragment KitFragment on Kit {
    id
    name
    primaryColor
    primaryTextColor
    secondaryColor
    secondaryTextColor
}
"""

team_full = """
fragment TeamFullFragment on Team {
    id
    name
    shortName
    availableSeasons
    country
    fifaId
    internationalName
}
"""

game = """
fragment GameFragment on Game {
    id
    awayTeam {
        ...TeamFragment
    }
    awayTeamKit {
        id
        name
        primaryColor
        primaryTextColor
        secondaryColor
        secondaryTextColor
    }
    homeTeam {
        ...TeamFragment
    }
    homeTeamKit {
        id
        name
        primaryColor
        primaryTextColor
        secondaryColor
        secondaryTextColor
    }
    competition {
        id
        name
    }
    complete
    date
    startPeriod1
    endPeriod1
    startPeriod2
    endPeriod2
    season
    stadium {
        id
        name
        pitches {
          length
          width
          grassType
        }
    }
    allRosters {
        ...RosterFragment
    }
    inallRosters {
        ...RosterFragment
    }
    rosters {
        ...RosterFragment
    }
    week
    homeTeamStartLeft
    homeTeamStartLeftExtraTime
}
"""

possession_event = """
fragment PossessionEventFragment on PossessionEvent {
    id
    ballCarryEvent {
      ...BallCarryEventFragment
    }
    challengeEvent {
      ...ChallengeEventFragment
    }
    clearanceEvent {
      ...ClearanceEventFragment
    }
    crossEvent {
      ...CrossEventFragment
    }
    defenders {
      ...DefenderFragment
    }
    duration
    endTime
    formattedGameClock
    fouls {
      ...FoulFragment
    }
    game {
      id
    }
    gameClock
    gameEvent {
      id
    }
    insertedAt
    lastInGameEvent
    passingEvent {
      ...PassingEventFragment
    }
    possessionEventType
    reboundEvent {
      ...ReboundEventFragment
    }
    shootingEvent {
      ...ShootingEventFragment
    }
    startTime
    updatedAt
    videoUrl
}
"""

ball_carry_event = """
fragment BallCarryEventFragment on BallCarryEvent {
    additionalChallenger1 {
      ...PlayerFragment
    }
    additionalChallenger2 {
      ...PlayerFragment
    }
    additionalChallenger3 {
      ...PlayerFragment
    }
    ballCarrierPlayer {
      ...PlayerFragment
    }
    ballCarryEndPointX
    ballCarryEndPointY
    ballCarryStartPointX
    ballCarryStartPointY
    ballCarryType
    carryType
    challengerPlayer {
      ...PlayerFragment
    }
    createsSpace
    defenderPlayer {
      ...PlayerFragment
    }
    defenderPointX
    defenderPointY
    dribbleEndPointX
    dribbleEndPointY
    dribbleOutcomeType
    dribbleStartPointX
    dribbleStartPointY
    dribbleType
    game {
      id
    }
    gameEvent {
        id
    }
    id
    linesBrokenType
    missedTouchPlayer {
      ...PlayerFragment
    }
    missedTouchPointX
    missedTouchPointY
    missedTouchType
    opportunityType
    possessionEvent {
        id
    }
    pressurePlayer {
      ...PlayerFragment
    }
    tackleAttemptPointX
    tackleAttemptPointY
    tackleAttemptType
    touchOutcomePlayer {
      ...PlayerFragment
    }
    touchOutcomeType
    touchPointX
    touchPointY
    touchType
    trickType
}
"""

challenge_event = """
fragment ChallengeEventFragment on ChallengeEvent {
additionalChallenger1 {
      ...PlayerFragment
    }
    additionalChallenger2 {
      ...PlayerFragment
    }
    additionalChallenger3 {
      ...PlayerFragment
    }
    ballCarrierPlayer {
      ...PlayerFragment
    }
    carryType
    challengeOutcomeType
    challengePointX
    challengePointY
    challengeType
    challengeWinnerPlayer {
      ...PlayerFragment
    }
    challengerAwayPlayer {
      ...PlayerFragment
    }
    challengerHomePlayer {
      ...PlayerFragment
    }
    challengerPlayer {
      ...PlayerFragment
    }
    createsSpace
    dribbleEndPointX
    dribbleEndPointY
    dribbleStartPointX
    dribbleStartPointY
    dribbleType
    game {
      id
    }
    gameEvent {
      id
    }
    id
    insertedAt
    keeperPlayer {
      ...PlayerFragment
    }
    linesBrokenType
    missedTouchPlayer {
      ...PlayerFragment
    }
    missedTouchPointX
    missedTouchPointY
    missedTouchType
    opportunityType
    possessionEvent {
      id
    }
    pressurePlayer {
      ...PlayerFragment
    }
    tackleAttemptPointX
    tackleAttemptPointY
    tackleAttemptType
    trickType
}
"""

clearance_event = """
fragment ClearanceEventFragment on ClearanceEvent {
    ballHeightType
    ballHighPointType
    blockerPlayer {
      ...PlayerFragment
    }
    clearanceBodyType
    clearanceEndPointX
    clearanceEndPointY
    clearanceOutcomeType
    clearancePlayer {
      ...PlayerFragment
    }
    clearancePointX
    clearancePointY
    clearanceStartPointX
    clearanceStartPointY
    createsSpace
    defenderPlayer {
      ...PlayerFragment
    }
    failedInterventionPlayer {
      ...PlayerFragment
    }
    failedInterventionPlayer1 {
      ...PlayerFragment
    }
    failedInterventionPlayer2 {
      ...PlayerFragment
    }
    failedInterventionPlayer3 {
      ...PlayerFragment
    }
    game {
      id
    }
    gameEvent {
      id
    }
    id
    keeperPlayer {
      ...PlayerFragment
    }
    missedTouchPlayer {
      ...PlayerFragment
    }
    missedTouchPointX
    missedTouchPointY
    missedTouchType
    opportunityType
    possessionEvent {
      id
    }
    pressurePlayer {
      ...PlayerFragment
    }
    pressureType
    shotInitialHeightType
    shotOutcomeType
}
"""

cross_event = """
fragment CrossEventFragment on CrossEvent {
    ballHeightType
    blockerPlayer {
      ...PlayerFragment
    }
    clearerPlayer {
      ...PlayerFragment
    }
    completeToPlayer {
      ...PlayerFragment
    }
    createsSpace
    crossEndPointX
    crossEndPointY
    crossHighPointType
    crossOutcomeType
    crossPointX
    crossPointY
    crossStartPointX
    crossStartPointY
    crossType
    crossZoneType
    crosserBodyType
    crosserPlayer {
      ...PlayerFragment
    }
    defenderBallHeightType
    defenderBodyType
    defenderPlayer {
      ...PlayerFragment
    }
    deflectionPointX
    deflectionPointY
    deflectorBodyType
    deflectorPlayer {
      ...PlayerFragment
    }
    failedInterventionPlayer {
      ...PlayerFragment
    }
    failedInterventionPlayer1 {
      ...PlayerFragment
    }
    failedInterventionPlayer2 {
      ...PlayerFragment
    }
    failedInterventionPlayer3 {
      ...PlayerFragment
    }
    failedInterventionPointX
    failedInterventionPointY
    game {
      id
    }
    gameEvent {
      id
    }
    goalkeeperPointX
    goalkeeperPointY
    id
    incompletionReasonType
    intendedTargetPlayer {
      ...PlayerFragment
    }
    intendedTargetPointX
    intendedTargetPointY
    keeperPlayer {
      ...PlayerFragment
    }
    linesBrokenType
    missedTouchPlayer {
      ...PlayerFragment
    }
    missedTouchPointX
    missedTouchPointY
    missedTouchType
    noLook
    opportunityType
    possessionEvent {
      id
    }
    pressurePlayer {
      ...PlayerFragment
    }
    pressureType
    receiverBallHeightType
    receiverBodyType
    receiverPointX
    receiverPointY
    secondIncompletionReasonType
    shotInitialHeightType
    shotOutcomeType
    targetZonePointX
    targetZonePointY
}
"""

defender = """
fragment DefenderFragment on Defender {
    defenderPlayer {
      ...PlayerFragment
    }
    defenderPointX
    defenderPointY
    game {
      id
    }
    gameEvent {
      id
    }
    id
    possessionEvent {
      id
    }
}
"""

foul = """
fragment FoulFragment on Foul {
    badCall
    correctDecision
    culpritPlayer {
      ...PlayerFragment
    }
    foulOutcomeType
    foulPointX
    foulPointY
    foulType
    game {
      id
    }
    gameEvent {
      id
    }
    id
    possessionEvent {
      id
    }
    potentialOffenseType
    sequence
    tacticalFoul
    var
    varCulpritPlayer {
      ...PlayerFragment
    }
    varOutcomeType
    varPotentialOffenseType
    varReasonType
    victimPlayer {
      ...PlayerFragment
    }
}
"""

passing_event = """
fragment PassingEventFragment on PassingEvent {
    ballHeightType
    blockerPlayer {
      ...PlayerFragment
    }
    clearerPlayer {
      ...PlayerFragment
    }
    createsSpace
    defenderBodyType
    defenderHeightType
    defenderPlayer {
      ...PlayerFragment
    }
    defenderPointX
    defenderPointY
    deflectionPointX
    deflectionPointY
    deflectorBodyType
    deflectorPlayer {
      ...PlayerFragment
    }
    failedInterventionPlayer {
      ...PlayerFragment
    }
    failedInterventionPlayer1 {
      ...PlayerFragment
    }
    failedInterventionPlayer2 {
      ...PlayerFragment
    }
    failedInterventionPlayer3 {
      ...PlayerFragment
    }
    failedInterventionPointX
    failedInterventionPointY
    game {
      id
    }
    gameEvent {
      id
    }
    goalkeeperPointX
    goalkeeperPointY
    id
    incompletionReasonType
    keeperPlayer {
      ...PlayerFragment
    }
    linesBrokenType
    missedTouchPlayer {
      ...PlayerFragment
    }
    missedTouchPointX
    missedTouchPointY
    missedTouchType
    noLook
    onTarget
    opportunityType
    outOfPlayPointX
    outOfPlayPointY
    passAccuracyType
    passBodyType
    passHighPointType
    passOutcomeType
    passPointX
    passPointY
    passType
    passerPlayer {
      ...PlayerFragment
    }
    possessionEvent {
      id
    }
    pressurePlayer {
      ...PlayerFragment
    }
    pressureType
    receiverBodyType
    receiverFacingType
    receiverHeightType
    receiverPlayer {
      ...PlayerFragment
    }
    receiverPointX
    receiverPointY
    secondIncompletionReasonType
    shotInitialHeightType
    shotOutcomeType
    targetFacingType
    targetPlayer {
      ...PlayerFragment
    }
    targetPointX
    targetPointY
}
"""

rebound_event = """
fragment ReboundEventFragment on ReboundEvent {
    blockerPlayer {
      ...PlayerFragment
    }
    defenderPlayer {
      ...PlayerFragment
    }
    game {
      id
    }
    gameEvent {
      id
    }
    id
    missedTouchPlayer {
      ...PlayerFragment
    }
    missedTouchPointX
    missedTouchPointY
    missedTouchType
    originateType
    possessionEvent {
      id
    }
    reboundBodyType
    reboundEndPointX
    reboundEndPointY
    reboundHeightType
    reboundHighPointType
    reboundOutcomeType
    reboundPointX
    reboundPointY
    reboundStartPointX
    reboundStartPointY
    rebounderPlayer {
      ...PlayerFragment
    }
    shotInitialHeightType
    shotOutcomeType
}
"""

shooting_event = """
fragment ShootingEventFragment on ShootingEvent {
    badParry
    ballHeightType
    ballMoving
    blockerPlayer {
      ...PlayerFragment
    }
    bodyMovementType
    clearerPlayer {
      ...PlayerFragment
    }
    createsSpace
    defenderPointX
    defenderPointY
    deflectionPointX
    deflectionPointY
    deflectorBodyType
    deflectorPlayer {
      ...PlayerFragment
    }
    failedInterventionPlayer {
      ...PlayerFragment
    }
    failedInterventionPlayer1 {
      ...PlayerFragment
    }
    failedInterventionPlayer2 {
      ...PlayerFragment
    }
    failedInterventionPlayer3 {
      ...PlayerFragment
    }
    game {
      id
    }
    gameEvent {
      id
    }
    goalLineEndPointX
    goalLineEndPointY
    goalLineStartPointX
    goalLineStartPointY
    goalkeeperPointX
    goalkeeperPointY
    id
    keeperTouchPointX
    keeperTouchPointY
    keeperTouchType
    missedTouchPlayer {
      ...PlayerFragment
    }
    missedTouchPointX
    missedTouchPointY
    missedTouchType
    noLook
    possessionEvent {
      id
    }
    pressurePlayer {
      ...PlayerFragment
    }
    pressureType
    saveHeightType
    savePointX
    savePointY
    saveReboundType
    saveable
    saverPlayer {
      ...PlayerFragment
    }
    shooterPlayer {
      ...PlayerFragment
    }
    shotBodyType
    shotInitialHeightType
    shotNatureType
    shotOutcomeType
    shotPointX
    shotPointY
    shotTargetPointX
    shotTargetPointY
    shotType
}
"""

game_events = """
fragment GameEventFragment on GameEvent {
    id
    bodyType
    defenderLocations {
      ...LocationFragment
    }
    duration
    endTime
    endType
    startPointX
    startPointY
    endPointX
    endPointY
    formattedGameClock
    gameClock
    gameEventType
    heightType
    initialTouchType
    offenderLocations {
      ...LocationFragment
    }
    otherPlayer {
      ...PlayerFragment
    }
    outType
    player {
      ...PlayerFragment
    }
    playerOff {
      ...PlayerFragment
    }
    playerOffType
    playerOn {
      ...PlayerFragment
    }
    possessionEvents {
      ...PossessionEventFragment
    }
    pressurePlayer {
      ...PlayerFragment
    }
    pressureType
    period
    setpieceType
    startTime
    subType
    team {
      ...TeamFragment
    }
    scoreValue
    touches
    touchesInBox
    videoUrl
}
"""

competition_season = """
fragment CompetitionSeasonFragment on CompetitionSeason {
  end
  season
  start
}
"""

competition_game = """
fragment CompetitionGameFragment on Game {
    id
    awayTeam {
        ...TeamFragment
    }
    homeTeam {
        ...TeamFragment
    }
    date
    season
    stadium {
        id
        name
        pitches { 
            length
            width
            grassType
        }
        noiseLevel
        seatingType
        stadiumType
        yearBuilt
        elevation
        capacity
    }
    week
}
"""

route = """
fragment RouteFragment on Route {
    confidence
    gameClockTime
    homeTeam
    isBall
    jersey
    playerId
    point
    visibility
}
"""

dribble_metrics = """
fragment DribbleMetricsFragment on PlayerDribbleMetricsReport {
    drb
    drbP90
    drbcmp
    drbcmpP90
    drbcmpPct
    gameId
    min
    playerId
    pos
    teamId
    week
}
"""


fragments = {
    'LocationFragment': location,
    'PlayerFragment': player_id,
    'PlayerFullFragment': player_full,
    'RosterFragment': roster,
    'TeamFragment': team,
    'GameFragment': game,
    'PossessionEventFragment': possession_event,
    'BallCarryEventFragment': ball_carry_event,
    'ChallengeEventFragment': challenge_event,
    'ClearanceEventFragment': clearance_event,
    'CrossEventFragment': cross_event,
    'DefenderFragment': defender,
    'FoulFragment': foul,
    'PassingEventFragment': passing_event,
    'ReboundEventFragment': rebound_event,
    'ShootingEventFragment': shooting_event,
    'GameEventFragment': game_events,
    'CompetitionSeasonFragment': competition_season,
    'CompetitionGameFragment': competition_game,
    'NationFragment': nation,
    'FederationFragment': federation,
    'ConfederationFragment': confederation,
    'RosterShortFragment': roster_short,
    'RouteFragment': route,
    'TeamFullFragment': team_full,
    'DribbleMetricsFragment': dribble_metrics,
}



from __future__ import annotations

from enum import Enum

from pydantic import BaseModel, Field


class BallCarryType(Enum):
    CARRY = 'C'
    DRIBBLE = 'D'
    TOUCH = 'T'


class BetterOptionType(Enum):
    BALL_CARRY = 'B'
    CROSS = 'C'
    HOLD = 'H'
    CLEARANCE = 'L'
    CONTINUE = 'O'
    PASS = 'P'
    SHOT = 'S'


class BodyMovementType(Enum):
    AWAY_FROM_GOAL = 'AG'
    LATERALLY = 'LA'
    STATIC = 'ST'
    TOWARDS_GOAL = 'TG'


class PFF_BodyType(Enum):
    BACK = 'BA'
    BOTTOM = 'BO'
    TWO_HAND_CATCH = 'CA'
    CHEST = 'CH'
    HEAD = 'HE'
    LEFT_FOOT = 'L'
    LEFT_ARM = 'LA'
    LEFT_BACK_HEEL = 'LB'
    LEFT_SHOULDER = 'LC'
    LEFT_HAND = 'LH'
    LEFT_KNEE = 'LK'
    LEFT_SHIN = 'LS'
    LEFT_THIGH = 'LT'
    TWO_HAND_PALM = 'PA'
    TWO_HAND_PUNCH = 'PU'
    RIGHT_FOOT = 'R'
    RIGHT_ARM = 'RA'
    RIGHT_BACK_HEEL = 'RB'
    RIGHT_SHOULDER = 'RC'
    RIGHT_HAND = 'RH'
    RIGHT_KNEE = 'RK'
    RIGHT_SHIN = 'RS'
    RIGHT_THIGH = 'RT'
    TWO_HANDS = 'TWOHANDS'
    VIDEO_MISSING = 'VM'


class CarryType(Enum):
    LINE_BREAK = 'B'
    CHANGE_OF_DIRECTION = 'C'
    DRIVE_WITH_INTENT = 'D'
    LONG_CARRY = 'L'


class PFF_ChallengeOutcomeType(Enum):
    DISTRIBUTION_DISRUPTED = 'B'
    FORCED_OUT_OF_PLAY = 'C'
    DISTRIBUTES_BALL = 'D'  # tem um passe/clearance/shot/cross na hora do challenge
    FOUL = 'F'
    SHIELDS_IN_PLAY = 'I'
    KEEPS_BALL_WITH_CONTACT = 'K'
    ROLLS = 'L'  # girar em volta do marcador (geralmente tem pressao)
    BEATS_MAN_LOSES_BALL = 'M'
    NO_WIN_KEEP_BALL = 'N'  # tentou driblar, nao conseguiu passar e manteve a posse
    OUT_OF_PLAY = 'O'
    PLAYER = 'P'  # a bola vai para um jogador que nao tava envolvido (ex: espirrou)
    RETAIN = 'R'
    SHIELDS_OUT_OF_PLAY = 'S'


class PFF_ChallengeType(Enum):
    AERIAL_DUEL = 'A'
    FROM_BEHIND = 'B'
    DRIBBLE = 'D'
    FIFTY = 'FIFTY'
    GK_SMOTHERS = 'G'
    SHIELDING = 'H'
    HAND_TACKLE = 'K'
    SLIDE_TACKLE = 'L'
    SHOULDER_TO_SHOULDER = 'S'
    STANDING_TACKLE = 'T'


class ClearanceOutcomeType(Enum):
    LUCKY_SHOT_AT_GOAL = 'A'
    BLOCK = 'B'
    LUCKY_SHOT_OWN_GOAL = 'D'
    OWN_POST = 'N'
    OUT_OF_PLAY = 'O'
    PLAYER = 'P'
    STOPPAGE = 'S'
    POST = 'W'


class CrossOutcomeType(Enum):
    BLOCKED = 'B'
    COMPLETE = 'C'
    DEFENSIVE_INTERCEPTION = 'D'
    LUCKY_SHOT_AT_GOAL = 'I'
    OUT_OF_PLAY = 'O'
    STOPPAGE = 'S'
    UNTOUCHED = 'U'


class CrossType(Enum):
    DRILLED = 'D'
    FLOATED = 'F'
    SWING_IN = 'I'
    SWING_OUT = 'O'
    PLACED = 'P'


class CrossZoneType(Enum):
    CENTRAL = 'C'
    FAR_POST = 'F'
    NEAR_POST = 'N'
    SIX_YARD_BOX = 'S'


class DribbleOutcomeType(Enum):
    KEEPS_BALL_WITH_CONTACT = 'B'
    FOUL = 'F'
    MISHIT = 'H'
    KEEPS_BALL = 'K'
    BEATS_MAN_LOSES_BALL = 'L'
    MISSED_FOUL = 'M'
    FORCED_OUT_OF_PLAY = 'O'
    SUCCESSFUL_TACKLE = 'S'


class DribbleType(Enum):
    BETWEEN_TWO_DEFENDERS = 'B'
    INSIDE = 'I'
    KNOCKS_IN_FRONT = 'K'
    OUTSIDE = 'O'
    TRICK = 'T'


class EndType(Enum):
    EXTRA_1 = 'F'
    FIRST = 'FIRST'
    GAME = 'G'
    EXTRA_2 = 'S'
    SECOND = 'SECOND'
    Z_TEST_9 = 'Z'


class FacingType(Enum):
    BACK_TO_GOAL = 'B'
    GOAL = 'G'
    LATERAL = 'L'


class PFF_FoulOutcomeType(Enum):
    NO_FOUL = 'F'
    NO_WARNING = 'N'
    RED_CARD = 'R'
    SECOND_YELLOW = 'S'
    WARNING = 'W'
    YELLOW_CARD = 'Y'


class PFF_FoulType(Enum):
    ADVANTAGE = 'A'
    INFRINGEMENT = 'I'
    MISSED_INFRINGEMENT = 'M'


class PFF_EventType(Enum):
    FIRST_HALF_KICKOFF = 'FIRSTKICKOFF'
    SECOND_HALF_KICKOFF = 'SECONDKICKOFF'
    EXTRA_1_KICKOFF = 'THIRDKICKOFF'
    EXTRA_2_KICKOFF = 'FOURTHKICKOFF'
    GAME_CLOCK_OBSERVATION = 'CLK'
    END_OF_HALF = 'END'
    GROUND = 'G'  # trave!!! travessao ou bandeira de escanteio. O evento eh vazio.
    PLAYER_OFF = 'OFF'
    PLAYER_ON = 'ON'
    POSSESSION = 'OTB'
    BALL_OUT_OF_PLAY = 'OUT'
    PAUSE_OF_GAME_TIME = 'PAU'
    SUB = 'SUB'
    VIDEO = 'VID'
    # Might not need below.
    FIRST_HALF_KICKOFF_OLD = '1KO'
    SECOND_HALF_KICKOFF_OLD = '2KO'
    EXTRA_1_KICKOFF_OLD = '3KO'
    EXTRA_2_KICKOFF_OLD = '4KO'


class HeightType(Enum):
    ABOVE_HEAD = 'A'
    GROUND = 'G'
    BETWEEN_WAIST_AND_HEAD = 'H'
    OFF_GROUND_BELOW_WAIST = 'L'
    VIDEO_MISSING = 'M'
    HALF_VOLLEY = 'V'  # Bola quica e ta subindo


class IncompletionReasonType(Enum):
    BEHIND = 'BH'
    BLOCKED = 'BL'
    CAUGHT = 'CA'
    DEFENSIVE_CONTACT = 'CO'
    DELIBERATE = 'DB'
    DEFENSIVE_CHALLENGE = 'DC'
    DEFLECTED = 'DF'
    DEFENDER_INTERCEPTION = 'DI'
    FOUL = 'FO'
    HIGH = 'HI'
    HIT_OFFICIAL = 'HO'
    IN_FRONT = 'IF'
    RECEIVER_LETS_BALL_RUN = 'LB'
    MISCOMMUNICATION = 'MC'
    MISS_HIT = 'MH'
    PASSER_SLIPPED = 'PS'
    RECEIVER_DIDNT_RETURN_TO_BALL = 'RB'
    RECEIVER_SLIPPED = 'RF'
    RECEIVER_MISSES_BALL = 'RM'
    RECEIVER_STOPPED = 'RS'
    REFEREE_IN_WAY = 'RW'
    SPECULATIVE = 'SP'
    UNDERHIT = 'UH'


class InitialTouchType(Enum):
    H2C_BAD = 'B'
    H2C_GOOD = 'G'
    MISCONTROL = 'M'
    STANDARD = 'S'


class LinesBrokenType(Enum):
    ATT = 'A'
    ATT_MID = 'AM'
    ATT_MID_DEF = 'AMD'
    DEF = 'D'
    MID = 'M'
    MID_DEF = 'MD'


class MissedTouchType(Enum):
    DUMMY = 'D'  # deixa a bola passar pra outro
    MISSED_TOUCH = 'M'
    SLIP = 'S'


class PFF_OpportunityType(Enum):
    CHANCE_CREATED = 'C'
    DANGEROUS_POSITION = 'D'
    HALF_CHANCE = 'H'
    SPACE_TO_CLEAR = 'L'
    NEGATIVE_CHANCE_CREATED = 'N'
    NEGATIVE_DANGEROUS_POSITION = 'P'
    SPACE_TO_CROSS = 'R'
    SPACE_TO_SHOOT = 'S'


class OriginateType(Enum):
    CORNER_FLAG = 'C'
    MISCELLANEOUS = 'M'
    PLAYER = 'P'
    POST = 'W'


class OutType(Enum):
    AWAY_SCORE = 'A'
    HOME_SCORE = 'H'
    TOUCH = 'T'
    WHISTLE = 'W'


class PassAccuracyType(Enum):
    CHECKS_MOVEMENT = 'C'
    HEAVY = 'H'
    LIGHT = 'L'
    PRECISE = 'P'
    REDIRECTS = 'R'
    STANDARD = 'S'


class PFF_PassOutcomeType(Enum):
    BLOCKED = 'B'
    COMPLETE = 'C'
    DEFENSIVE_INTERCEPTION = 'D'
    LUCKY_SHOT_OWN_GOAL = 'G'
    LUCKY_SHOT_GOAL = 'I'
    OUT_OF_PLAY = 'O'
    STOPPAGE = 'S'


class PFF_PassType(Enum):
    CUTBACK = 'B'
    CREATE_CONTEST = 'C'
    FLICK_ON = 'F'
    LONG_PASS = 'L'
    MISS_HIT = 'M'
    BALL_OVER_THE_TOP = 'O'
    STANDARD_PASS = 'S'
    THROUGH_BALL = 'T'


class PlayerOffType(Enum):
    INJURY = 'I'
    RED_CARD = 'R'
    YELLOW_CARD = 'Y'


class PositionGroupType(Enum):
    ATTACK_MID = 'AM'
    CENTER_FORWARD = 'CF'
    CENTER_MID = 'CM'
    DEFENDER = 'D'
    DEFENSIVE_MID = 'DM'
    FORWARD = 'F'
    GK = 'GK'
    LEFT_BACK = 'LB'
    LEFT_CENTER_BACK = 'LCB'
    LEFT_MID = 'LM'
    LEFT_WINGER = 'LW'
    LEFT_WING_BACK = 'LWB'
    MIDFIELDER = 'M'
    MID_CENTER_BACK = 'MCB'
    RIGHT_BACK = 'RB'
    RIGHT_CENTER_BACK = 'RCB'
    CENTER_BACK = 'CB'
    RIGHT_MID = 'RM'
    RIGHT_WINGER = 'RW'
    RIGHT_WING_BACK = 'RWB'


class PFF_PossessionEventType(Enum):
    BALL_CARRY = 'BC'
    CHALLENGE = 'CH'
    CLEARANCE = 'CL'
    CROSS = 'CR'
    PASS = 'PA'
    REBOUND = 'RE'
    SHOT = 'SH'


class PFF_PotentialOffenseType(Enum):
    DISSENT = 'D'
    OFF_THE_BALL = 'F'
    HAND_BALL = 'H'
    ON_THE_BALL = 'N'
    OFFSIDE = 'O'
    TECHNICAL = 'T'
    DIVA = 'V'


class PressureType(Enum):
    ATTEMPTED = 'A'
    PASSING_LANE = 'L'
    PRESSURED = 'P'


# Rebound: a bola bateu no jogador
class ReboundOutcomeType(Enum):
    LUCKY_SHOT_GOAL = 'A'
    LUCKY_SHOT_OWN_GOAL = 'D'
    PLAYER = 'P'
    RETAIN = 'R'
    OUT_OF_TOUCH = 'T'


class SaveReboundType(Enum):
    CROSSBAR = 'CB'
    LEFT_BEHIND_GOAL = 'GL'
    RIGHT_BEHIND_GOAL = 'GR'
    LEFT_BEHIND_GOAL_HIGH = 'HL'
    RIGHT_BEHIND_GOAL_HIGH = 'HR'
    LEFT_SIX_YARD_BOX = 'L6'
    LEFT_AREA = 'LA'
    LEFT_OUT_OF_BOX = 'LO'
    LEFT_POST = 'LP'
    MIDDLE_SIX_YARD_BOX = 'M6'
    MIDDLE_AREA = 'MA'
    MIDDLE_OUT_OF_BOX = 'MO'
    CROSSBAR_OVER = 'OC'
    RIGHT_SIX_YARD_BOX = 'R6'
    RIGHT_AREA = 'RA'
    RIGHT_OUT_OF_BOX = 'RO'
    RIGHT_POST = 'RP'


class PFF_SetpieceType(Enum):
    CORNER = 'C'
    DROP_BALL = 'D'
    FREE_KICK = 'F'
    GOAL_KICK = 'G'
    KICKOFF = 'K'
    PENALTY = 'P'
    THROW_IN = 'T'


class ShotHeightType(Enum):
    BOTTOM_THIRD = 'BOTTOMTHIRD'
    CROSSBAR = 'C'
    SHORT = 'F'
    GROUND = 'G'
    MIDDLE_THIRD = 'MIDDLETHIRD'
    CROSSBAR_NARROW_OVER = 'N'
    OVER = 'O'
    TOP_THIRD = 'TOPTHIRD'
    CROSSBAR_NARROW_UNDER = 'U'


class ShotNatureType(Enum):
    PLACEMENT = 'A'
    FLICK = 'F'
    LACES = 'L'
    POWER = 'P'
    SCUFF = 'S'
    TOE_PUNT = 'T'


class PFF_ShotOutcomeType(Enum):
    ON_TARGET_BLOCK = 'B'
    OFF_TARGET_BLOCK = 'C'
    SAVE_OFF_TARGET = 'F'
    GOAL = 'G'
    GOALLINE_CLEARANCE = 'L'
    OFF_TARGET = 'O'
    ON_TARGET = 'S'


class PFF_ShotType(Enum):
    BICYCLE = 'B'
    DIVING = 'D'
    SIDE_FOOT = 'F'
    SLIDING = 'I'
    LOB = 'L'
    OUTSIDE_FOOT = 'O'
    STANDARD = 'S'
    STUDS = 'T'
    VOLLEY = 'V'


class PFF_StadiumGrassType(Enum):
    ASTRO_TURF = 'A'
    FIELD_TURF = 'F'
    REAL = 'R'
    NATURAL = 'N'


class PFF_StadiumType(Enum):
    CONVERSION = 'C'
    DOMED = 'D'
    INDOOR = 'I'
    OUTDOOR = 'O'


class SubType(Enum):
    BLOOD = 'B'
    SIN_BIN_COVER = 'C'
    HEAD = 'H'
    RETURN_FROM_HIA = 'R'
    STANDARD = 'S'


class TackleAttemptType(Enum):
    DELIBERATE_FOUL = 'D'
    NO_TACKLE_FAKE_EVENT = 'F'
    GO_FOR_BALL = 'G'
    NO_TACKLE = 'T'


class TouchOutcomeType(Enum):
    CHALLENGE = 'C'
    GOAL = 'G'
    OUT_OF_PLAY = 'O'
    PLAYER = 'P'
    OWN_GOAL = 'W'


class TouchType(Enum):
    BALL_IN_HAND = 'B'
    FAILED_CROSS = 'C'
    HAND_BALL = 'D'
    FAILED_TRAP = 'F'
    FAILED_CATCH = 'G'
    HEAVY_TOUCH = 'H'
    FAILED_CLEARANCE = 'L'
    FAILED_PASS = 'P'
    FAILED_SHOT = 'S'
    TAKE_OVER = 'T'


class VarReasonType(Enum):
    MISSED = 'I'
    OVERTURN = 'O'


class VideoAngleType(Enum):
    BAD_ANGLE = 'B'
    MISSING = 'M'


class _ID(BaseModel):
    id: str | None = None


class PFF_BallCarryEvent(BaseModel):
    id: str | None = None
    additionalChallenger1: Player | None = None
    additionalChallenger2: Player | None = None
    additionalChallenger3: Player | None = None
    ballCarrierPlayer: Player | None = None
    ballCarryEndPointX: float | None = None
    ballCarryEndPointY: float | None = None
    ballCarryStartPointX: float | None = None
    ballCarryStartPointY: float | None = None
    ballCarryType: BallCarryType | None = None
    carryType: CarryType | None = None
    challengerPlayer: Player | None = None
    createsSpace: bool | None = None
    defenderPlayer: Player | None = None
    defenderPointX: float | None = None
    defenderPointY: float | None = None
    dribbleEndPointX: float | None = None
    dribbleEndPointY: float | None = None
    dribbleOutcomeType: DribbleOutcomeType | None = None
    dribbleStartPointX: float | None = None
    dribbleStartPointY: float | None = None
    dribbleType: DribbleType | None = None
    game: PFF_Game | None = None
    gameEvent: PFF_Event | None = None
    linesBrokenType: LinesBrokenType | None = None
    opportunityType: PFF_OpportunityType | None = None
    period: str | int | None = None
    possessionEvent: PFF_PossessionEvent | None = None
    pressurePlayer: Player | None = None
    tackleAttemptPointX: float | None = None
    tackleAttemptPointY: float | None = None
    tackleAttemptType: TackleAttemptType | None = None
    touchOutcomePlayer: Player | None = None
    touchOutcomeType: TouchOutcomeType | None = None
    touchPointX: float | None = None
    touchPointY: float | None = None
    touchType: TouchType | None = None
    trickType: str | None = None


class CacheStats(BaseModel):
    hitRate: float | None = None
    name: str | None = None


class PFF_ChallengeEvent(BaseModel):
    id: str | None = None
    additionalChallenger1: Player | None = None
    additionalChallenger2: Player | None = None
    additionalChallenger3: Player | None = None
    ballCarrierPlayer: Player | None = None
    challengeOutcomeType: PFF_ChallengeOutcomeType | None = None
    challengePointX: float | None = None
    challengePointY: float | None = None
    challengeType: PFF_ChallengeType | None = None
    challengeWinnerPlayer: Player | None = None
    challengerAwayPlayer: Player | None = None
    challengerHomePlayer: Player | None = None
    challengerPlayer: Player | None = None
    createsSpace: bool | None = None
    dribbleEndPointX: float | None = None
    dribbleEndPointY: float | None = None
    dribbleStartPointX: float | None = None
    dribbleStartPointY: float | None = None
    dribbleType: DribbleType | None = None
    game: PFF_Game | None = None
    gameEvent: PFF_Event | None = None
    keeperPlayer: Player | None = None
    linesBrokenType: LinesBrokenType | None = None
    opportunityType: PFF_OpportunityType | None = None
    period: str | None = None
    possessionEvent: PFF_PossessionEvent | None = None
    pressurePlayer: Player | None = None
    tackleAttemptPointX: float | None = None
    tackleAttemptPointY: float | None = None
    tackleAttemptType: TackleAttemptType | None = None
    trickType: str | None = None


class PFF_ClearanceEvent(BaseModel):
    ballHeightType: str | None = None
    ballHighPointType: str | None = None
    blockerPlayer: Player | None = None
    clearanceBodyType: PFF_BodyType
    clearanceEndPointX: float | None = None
    clearanceEndPointY: float | None = None
    clearanceOutcomeType: ClearanceOutcomeType
    clearancePlayer: Player
    clearancePointX: float | None = None
    clearancePointY: float | None = None
    clearanceStartPointX: float | None = None
    clearanceStartPointY: float | None = None
    createsSpace: bool | None = None
    defenderPlayer: Player | None = None
    failedInterventionPlayer: Player | None = None
    failedInterventionPlayer1: Player | None = None
    failedInterventionPlayer2: Player | None = None
    failedInterventionPlayer3: Player | None = None
    game: PFF_Game | None = None
    gameEvent: PFF_Event | None = None
    id: str | None = None
    keeperPlayer: Player | None = None
    opportunityType: PFF_OpportunityType | None = None
    period: str | None = None
    possessionEvent: PFF_PossessionEvent | None = None
    pressurePlayer: Player | None = None
    pressureType: PressureType | None = None
    shotInitialHeightType: ShotHeightType | None = None
    shotOutcomeType: PFF_ShotOutcomeType | None = None


class PFF_Competition(BaseModel):
    availableSeasons: list[PFF_CompetitionSeason] | None = Field(default_factory=list)
    id: str
    games: list[PFF_Game] | None = Field(default_factory=list)
    name: str | None = None
    seasonGames: list[PFF_Game] | None = Field(default_factory=list)
    teams: list[PFF_Team] | None = Field(default_factory=list)


class PFF_CompetitionSeason(BaseModel):
    season: str | None = None
    start: str | None = None
    end: str | None = None


class Confederation(BaseModel):
    id: str | None = None
    abbreviation: str | None = None
    name: str | None = None


class PFF_CrossEvent(BaseModel):
    ballHeightType: HeightType | None = None
    blockerPlayer: Player | None = None
    clearerPlayer: Player | None = None
    completeToPlayer: Player | None = None
    createsSpace: bool | None = None
    crossEndPointX: float | None = None
    crossEndPointY: float | None = None
    crossHighPointType: str | None = None
    crossOutcomeType: CrossOutcomeType | None = None
    crossPointX: float | None = None
    crossPointY: float | None = None
    crossStartPointX: float | None = None
    crossStartPointY: float | None = None
    crossType: CrossType | None = None
    crossZoneType: CrossZoneType | None = None
    crosserBodyType: PFF_BodyType | None = None
    crosserPlayer: Player | None = None
    defenderBallHeightType: HeightType | None = None
    defenderBodyType: PFF_BodyType | None = None
    defenderPlayer: Player | None = None
    deflectionPointX: float | None = None
    deflectionPointY: float | None = None
    deflectorBodyType: PFF_BodyType | None = None
    deflectorPlayer: Player | None = None
    failedInterventionPlayer: Player | None = None
    failedInterventionPlayer1: Player | None = None
    failedInterventionPlayer2: Player | None = None
    failedInterventionPlayer3: Player | None = None
    failedInterventionPointX: float | None = None
    failedInterventionPointY: float | None = None
    game: PFF_Game | None = None
    gameEvent: PFF_Event | None = None
    goalkeeperPointX: float | None = None
    goalkeeperPointY: float | None = None
    id: str | None = None
    incompletionReasonType: IncompletionReasonType | None = None
    intendedTargetPlayer: Player | None = None
    intendedTargetPointX: float | None = None
    intendedTargetPointY: float | None = None
    keeperPlayer: Player | None = None
    linesBrokenType: LinesBrokenType | None = None
    opportunityType: PFF_OpportunityType | None = None
    period: str | None = None
    possessionEvent: PFF_PossessionEvent | None = None
    pressurePlayer: Player | None = None
    pressureType: PressureType | None = None
    receiverBallHeightType: HeightType | None = None
    receiverBodyType: PFF_BodyType | None = None
    receiverPointX: float | None = None
    receiverPointY: float | None = None
    secondIncompletionReasonType: IncompletionReasonType | None = None
    shotInitialHeightType: ShotHeightType | None = None
    shotOutcomeType: PFF_ShotOutcomeType | None = None
    targetZonePointX: float | None = None
    targetZonePointY: float | None = None


class Defender(BaseModel):
    id: str | None = None
    defenderPlayer: Player | None = None
    defenderPointX: float | None = None
    defenderPointY: float | None = None
    game: PFF_Game | None = None
    gameEvent: PFF_Event | None = None
    possessionEvent: PFF_PossessionEvent | None = None


class Federation(BaseModel):
    id: str | None = None
    name: str | None = None
    englishName: str | None = None
    abbreviation: str | None = None
    confederation: Confederation | None = None
    country: str | None = None


class PFF_Foul(BaseModel):
    id: str
    badCall: bool | None = None
    correctDecision: bool | None = None
    culpritPlayer: Player | None = None
    foulOutcomeType: PFF_FoulOutcomeType | None = None
    foulPointX: float | None = None
    foulPointY: float | None = None
    foulType: PFF_FoulType | None = None
    game: PFF_Game | None = None
    gameEvent: PFF_Event | None = None
    possessionEvent: PFF_PossessionEvent | None = None
    potentialOffenseType: PFF_PotentialOffenseType | None = None
    sequence: int | None = None
    tacticalFoul: bool | None = None
    var: bool | None = None
    varCulpritPlayer: Player | None = None
    varOutcomeType: PFF_FoulOutcomeType | None = (
        None  # VAR OutcomeType | PFF_FoulOutcomeType
    )
    varPotentialOffenseType: PFF_PotentialOffenseType | None = None
    varReasonType: VarReasonType | None = None
    victimPlayer: Player | None = None


class Location(BaseModel):
    id: str | None = None
    ballCarryEvent: PFF_BallCarryEvent | None = None
    challengeEvent: PFF_ChallengeEvent | None = None
    clearanceEvent: PFF_ClearanceEvent | None = None
    crossEvent: PFF_CrossEvent | None = None
    gameEvent: PFF_Event | None = None
    name: str | None = None
    passingEvent: PFF_PassingEvent | None = None
    possessionEvent: PFF_PossessionEvent | None = None
    reboundEvent: PFF_ReboundEvent | None = None
    shootingEvent: PFF_ShootingEvent | None = None
    x: float | None = None
    y: float | None = None


class Nation(BaseModel):
    country: str | None = None
    federation: Federation | None = None
    fifaCode: str | None = None
    id: str | None = None
    iocCode: str | None = None


class PFF_PassingEvent(BaseModel):
    ballHeightType: HeightType | None = None
    blockerPlayer: Player | None = None
    clearerPlayer: Player | None = None
    defenderBodyType: PFF_BodyType | None = None
    defenderHeightType: HeightType | None = None
    defenderPlayer: Player | None = None
    defenderPointX: float | None = None
    defenderPointY: float | None = None
    deflectionPointX: float | None = None
    deflectionPointY: float | None = None
    deflectorBodyType: PFF_BodyType | None = None
    deflectorPlayer: Player | None = None
    failedInterventionPlayer: Player | None = None
    failedInterventionPlayer1: Player | None = None
    failedInterventionPlayer2: Player | None = None
    failedInterventionPlayer3: Player | None = None
    failedInterventionPointX: float | None = None
    failedInterventionPointY: float | None = None
    game: PFF_Game | None = None
    gameEvent: PFF_Event | None = None
    goalkeeperPointX: float | None = None
    goalkeeperPointY: float | None = None
    id: str | None = None
    incompletionReasonType: IncompletionReasonType | None = None
    keeperPlayer: Player | None = None
    linesBrokenType: LinesBrokenType | None = None
    missedTouchPlayer: Player | None = None
    missedTouchPointX: float | None = None
    missedTouchPointY: float | None = None
    missedTouchType: MissedTouchType | None = None
    noLook: bool | None = None
    onTarget: bool | None = None
    opportunityType: PFF_OpportunityType | None = None
    outOfPlayPointX: float | None = None
    outOfPlayPointY: float | None = None
    passAccuracyType: PassAccuracyType | None = None
    passBodyType: PFF_BodyType
    passHighPointType: HeightType | None = None
    passOutcomeType: PFF_PassOutcomeType
    passPointX: float | None = None
    passPointY: float | None = None
    passType: PFF_PassType | None = None
    passerPlayer: Player
    period: str | None = None
    possessionEvent: PFF_PossessionEvent | None = None
    pressurePlayer: Player | None = None
    pressureType: PressureType | None = None
    receiverBodyType: PFF_BodyType | None = None
    receiverFacingType: FacingType | None = None
    receiverHeightType: HeightType | None = None
    receiverPlayer: Player | None = None
    receiverPointX: float | None = None
    receiverPointY: float | None = None
    secondIncompletionReasonType: IncompletionReasonType | None = None
    shotInitialHeightType: ShotHeightType | None = None
    shotOutcomeType: PFF_ShotOutcomeType | None = None
    targetFacingType: FacingType | None = None
    targetPlayer: Player | None = None
    targetPointX: float | None = None
    targetPointY: float | None = None


class Player(BaseModel):
    id: str
    dob: str | None = None
    firstName: str | None = None
    gender: str | None = None
    height: float | None = None  # in cm
    lastName: str | None = None
    nickname: str | None = None
    preferredFoot: str | None = None
    weight: float | None = None
    positionGroupType: PositionGroupType | None = None
    nationality: Nation | None = None
    secondNationality: Nation | None = None
    countryOfBirth: Nation | None = None
    # rosters: list[PFF_Roster] | None = Field(default_factory=list)


class PFF_PossessionEvent(BaseModel):
    id: str
    ballCarryEvent: PFF_BallCarryEvent | None = None
    challengeEvent: PFF_ChallengeEvent | None = None
    clearanceEvent: PFF_ClearanceEvent | None = None
    crossEvent: PFF_CrossEvent | None = None
    defenders: list[Defender] | None = Field(default_factory=list)
    duration: float | None = None
    endTime: float | None = None
    formattedGameClock: str | None = None
    fouls: list[PFF_Foul] | None = Field(default_factory=list)
    game: PFF_Game | None = None
    gameClock: float | None = None
    gameEvent: PFF_Event | None = None
    lastInGameEvent: int | None = None
    passingEvent: PFF_PassingEvent | None = None
    period: str | None = None
    possessionEventType: PFF_PossessionEventType | None = None
    reboundEvent: PFF_ReboundEvent | None = None
    shootingEvent: PFF_ShootingEvent | None = None
    startTime: float | None = None
    videoUrl: str | None = None  # Field(default=None, exclude=True)

    def __repr__(self) -> str:
        return f'{self.possessionEventType}={self.id}'


class PFF_ReboundEvent(BaseModel):
    id: str | None = None
    game: PFF_Game | None = None
    gameEvent: PFF_Event | None = None
    period: str | None = None
    originateType: OriginateType | None = None
    possessionEvent: PFF_PossessionEvent | None = None
    reboundBodyType: PFF_BodyType | None = None
    reboundEndPointX: float | None = None
    reboundEndPointY: float | None = None
    reboundHeightType: HeightType | None = None
    reboundHighPointType: HeightType | None = None
    reboundOutcomeType: ReboundOutcomeType | None = None
    reboundPointX: float | None = None
    reboundPointY: float | None = None
    reboundStartPointX: float | None = None
    reboundStartPointY: float | None = None
    rebounderPlayer: Player | None = None
    shotInitialHeightType: ShotHeightType | None = None
    shotOutcomeType: PFF_ShotOutcomeType | None = None


class PFF_Roster(BaseModel):
    id: str | None = None
    game: dict | None = None
    player: Player
    positionGroupType: PositionGroupType
    shirtNumber: str | int
    started: bool | None = None
    team: PFF_Team


class PFF_ShootingEvent(BaseModel):
    advantageType: str | None = None
    badParry: bool | None = None
    ballHeightType: HeightType | None = None
    ballMoving: bool | None = None
    betterOption: str | None = None
    betterOptionPlayer: Player | None = None
    betterOptionTime: str | None = None
    betterOptionType: BetterOptionType | None = None
    blockerPlayer: Player | None = None
    bodyMovementType: BodyMovementType | None = None
    clearerPlayer: Player | None = None
    createsSpace: bool | None = None
    defenderPointX: float | None = None
    defenderPointY: float | None = None
    deflectionPointX: float | None = None
    deflectionPointY: float | None = None
    deflectorBodyType: PFF_BodyType | None = None
    deflectorPlayer: Player | None = None
    goalLineEndPointX: float | None = None
    goalLineEndPointY: float | None = None
    goalLineStartPointX: float | None = None
    goalLineStartPointY: float | None = None
    goalkeeperPointX: float | None = None
    goalkeeperPointY: float | None = None
    id: str | None = None
    keeperTouchPointX: float | None = None
    keeperTouchPointY: float | None = None
    keeperTouchType: PFF_BodyType | None = None
    failedInterventionPlayer: Player | None = None
    missedTouchPlayer: Player | None = None
    noLook: bool | None = None
    period: str | None = None
    possessionEvent: PFF_PossessionEvent | None = None
    pressurePlayer: Player | None = None
    pressureType: PressureType | None = None
    saveHeightType: ShotHeightType | None = None
    savePointX: float | None = None
    savePointY: float | None = None
    saveReboundType: SaveReboundType | None = None
    saverPlayer: Player | None = None
    shooterPlayer: Player
    shotBodyType: PFF_BodyType
    shotInitialHeightType: ShotHeightType | None = None
    shotNatureType: ShotNatureType | None = None
    shotOutcomeType: PFF_ShotOutcomeType
    shotPointX: float | None = None
    shotPointY: float | None = None
    shotTargetPointX: float | None = None
    shotTargetPointY: float | None = None
    shotType: PFF_ShotType


class PFF_Pitch(BaseModel):
    length: float
    width: float
    grassType: PFF_StadiumGrassType


class PFF_Stadium(BaseModel):
    id: str
    name: str
    pitches: list[PFF_Pitch] = Field(default_factory=list)


class PFF_Team(BaseModel):
    id: str
    name: str | None = None
    shortName: str | None = None


class PFF_Kit(BaseModel):
    id: str
    name: str
    primaryColor: str
    primaryTextColor: str
    secondaryColor: str
    secondaryTextColor: str


class PFF_Game(BaseModel):
    id: str
    allRosters: list[PFF_Roster] = Field(default_factory=list)
    awayTeam: PFF_Team | None = None
    homeTeam: PFF_Team | None = None
    awayTeamKit: PFF_Kit | None = None
    homeTeamKit: PFF_Kit | None = None
    competition: PFF_Competition | None = None
    complete: bool | None = None
    date: str | None = None
    endPeriod1: float | None = None
    endPeriod2: float | None = None
    gameEvents: list[PFF_Event] | None = Field(default_factory=list)
    rosters: list[PFF_Roster] | None = Field(default_factory=list)
    home_started_left: bool | None = Field(None, alias='homeTeamStartLeft')
    home_started_left_et: bool | None = Field(None, alias='homeTeamStartLeftExtraTime')
    season: str | int | None = None
    stadium: PFF_Stadium | None = None
    startPeriod1: float | None = None
    startPeriod2: float | None = None
    week: int | None = None

    def get_by_event_type(self, event_type: PFF_EventType) -> list[PFF_Event]:
        if not self.gameEvents:
            return []

        return [evt for evt in self.gameEvents if evt.gameEventType == event_type]

    def get_by_possession_event_type(
        self, possession_event_type: PFF_PossessionEventType
    ) -> list[PFF_Event]:
        if not self.gameEvents:
            return []

        return [
            evt
            for evt in self.gameEvents
            if evt.possessionEvents
            and any(
                (p_evt.possessionEventType == possession_event_type)
                for p_evt in evt.possessionEvents
            )
        ]


class PFF_Event(BaseModel):
    # model_config = ConfigDict(use_enum_values=False)
    id: str
    bodyType: PFF_BodyType | None = None
    defenderLocations: list[Location] | None = Field(default_factory=list)
    duration: float | None = None
    endPointX: float | None = None
    endPointY: float | None = None
    endTime: float | None = None
    endType: EndType | None = None
    formattedGameClock: str | None = None
    game: PFF_Game | None = None
    gameClock: float | None = None
    gameEventType: PFF_EventType | None = None
    heightType: HeightType | None = None
    initialTouchType: InitialTouchType | None = None
    offenderLocations: list[Location] | None = Field(default_factory=list)
    otherPlayer: Player | None = None
    outType: OutType | None = None
    period: str | int | None = None
    player: Player | None = None
    playerOff: Player | None = None
    playerOffType: PlayerOffType | None = None
    playerOn: Player | None = None
    possessionEvents: list[PFF_PossessionEvent] = Field(default_factory=list)
    pressurePlayer: Player | None = None
    pressureType: PressureType | None = None
    setpieceType: PFF_SetpieceType | None = None
    scoreValue: int | None = None
    startPointX: float | None = None
    startPointY: float | None = None
    startTime: float | None = None
    subType: SubType | None = None
    team: PFF_Team | None = None
    touches: int | None = None
    touchesInBox: int | None = None

    def get_by_possession_event_type(
        self, possession_event_type: PFF_PossessionEventType
    ) -> list[PFF_PossessionEvent]:
        if not self.possessionEvents:
            return []

        return [
            evt
            for evt in self.possessionEvents
            if evt.possessionEventType == possession_event_type
        ]

    def get_by_possession_event_id(
        self, possession_event_id: str
    ) -> list[PFF_PossessionEvent]:
        if not self.possessionEvents:
            return []

        return [evt for evt in self.possessionEvents if evt.id == possession_event_id]

    def get_possession_events(self) -> list[PFF_PossessionEvent]:
        if not self.possessionEvents:
            return []

        return self.possessionEvents

    def __repr__(self) -> str:
        return f'PFF_Event({self.id}) {self.possessionEvents}'


class PassingEvent(PFF_PassingEvent): ...


class PassOutcomeType(Enum):
    BLOCKED = 'B'
    COMPLETE = 'C'
    DEFENSIVE_INTERCEPTION = 'D'
    LUCKY_SHOT_OWN_GOAL = 'G'
    LUCKY_SHOT_GOAL = 'I'
    OUT_OF_PLAY = 'O'
    STOPPAGE = 'S'


class SetpieceType(Enum):
    CORNER = 'C'
    DROP_BALL = 'D'
    FREE_KICK = 'F'
    GOAL_KICK = 'G'
    KICKOFF = 'K'
    PENALTY = 'P'
    THROW_IN = 'T'


class PotentialOffenseType(Enum):
    DISSENT = 'D'
    OFF_THE_BALL = 'F'
    HAND_BALL = 'H'
    ON_THE_BALL = 'N'
    OFFSIDE = 'O'
    TECHNICAL = 'T'
    DIVA = 'V'


class PassType(Enum):
    CUTBACK = 'B'
    CREATE_CONTEST = 'C'
    FLICK_ON = 'F'
    LONG_PASS = 'L'
    MISS_HIT = 'M'
    BALL_OVER_THE_TOP = 'O'
    STANDARD_PASS = 'S'
    THROUGH_BALL = 'T'


class OpportunityType(Enum):
    CHANCE_CREATED = 'C'
    DANGEROUS_POSITION = 'D'
    HALF_CHANCE = 'H'
    SPACE_TO_CLEAR = 'L'
    NEGATIVE_CHANCE_CREATED = 'N'
    NEGATIVE_DANGEROUS_POSITION = 'P'
    SPACE_TO_CROSS = 'R'
    SPACE_TO_SHOOT = 'S'


class ShootingEvent(PFF_ShootingEvent): ...


class ShotOutcomeType(Enum):
    ON_TARGET_BLOCK = 'B'
    OFF_TARGET_BLOCK = 'C'
    SAVE_OFF_TARGET = 'F'
    GOAL = 'G'
    GOALLINE_CLEARANCE = 'L'
    OFF_TARGET = 'O'
    ON_TARGET = 'S'


class ShotType(Enum):
    BICYCLE = 'B'
    DIVING = 'D'
    SIDE_FOOT = 'F'
    SLIDING = 'I'
    LOB = 'L'
    OUTSIDE_FOOT = 'O'
    STANDARD = 'S'
    STUDS = 'T'
    VOLLEY = 'V'



from gql.transport.exceptions import TransportServerError

from ..providers.pff.api import auth

# TODO: Rename module to something else more descriptive, meta PFF FC API


def validate_api_key(api_key: str | None = None, *, api_url: str | None = None) -> bool:
    """Validates the provided PFF API key.

    :param api_key: The API key to validate.
    :return: True if the API key is valid, False otherwise.
    """
    try:
        response = auth.account_active(pff_api_key=api_key, pff_api_url=api_url)
        return response['accountActive']
    except TransportServerError:
        return False


def api_version(*, api_url: str | None = None, api_key: str | None = None) -> dict:
    """Retrieves the version of the API.

    :param api_url: The API URL to connect to.
    :param api_key: The API
    :return: The version of the API.
    """
    return auth.version(pff_api_key=api_key, pff_api_url=api_url)



from datetime import datetime
from enum import Enum

from pydantic import BaseModel, ConfigDict, Field


class JerseyConfidence(Enum):
    HIGH = 'HIGH'
    MEDIUM = 'MEDIUM'
    LOW = 'LOW'


class Visibility(Enum):
    VISIBLE = 'VISIBLE'
    ESTIMATED = 'ESTIMATED'


class Frame_Player(BaseModel):
    shirt: int = Field(alias='jerseyNum')
    shirt_confidence: JerseyConfidence = Field(alias='confidence')
    visibility: Visibility
    x: float
    y: float


class Ball(BaseModel):
    visibility: Visibility
    x: float | None
    y: float | None
    z: float | None


class PFF_Frame_GameEventType(Enum):
    FIRST_KICK_OFF = 'FIRSTKICKOFF'
    SECOND_KICK_OFF = 'SECONDKICKOFF'
    THIRD_KICK_OFF = 'THIRDKICKOFF'
    FOURTH_KICK_OFF = 'FOURTHKICKOFF'
    FIRST_HALF_KICKOFF = '1KO'
    SECOND_HALF_KICKOFF = '2KO'
    END_OF_HALF = 'END'
    PBC_IN_PLAY = 'G'
    PLAYER_ON = 'ON'
    PLAYER_OFF = 'OFF'
    ON_THE_BALL = 'OTB'
    OUT_OF_PLAY = 'OUT'
    SUB = 'SUB'
    VIDEO_MISSING = 'VID'
    CLOCK = 'CLK'  # TODO: Check if this is correct


class SetPieceType(Enum):
    CORNER = 'C'
    DROP_BALL = 'D'
    FREE_KICK = 'F'
    GOAL_KICK = 'G'
    KICK_OFF = 'K'
    PENALTY = 'P'
    THROW_IN = 'T'


class Frame_Event(BaseModel):
    game_id: int
    game_event_type: PFF_Frame_GameEventType
    competition_id: int | None = None
    season: str | int | None = None
    clock: str | None = Field(alias='formatted_game_clock')
    player_id: int | None
    team_id: int | None
    setpiece_type: SetPieceType | None = None
    touches: int | None = Field(
        None, description='Number of touches taken by player on-the-ball'
    )
    touches_in_box: int | None = Field(
        None, description='Number of touches taken by player on-the-ball in the box'
    )
    start_time_seconds: float = Field(alias='start_time')
    end_time_seconds: float | None = Field(alias='end_time', default=None)
    duration_seconds: float | None = Field(alias='duration', default=None)
    video_missing: bool | None = Field(default=False)
    inserted_at: datetime
    updated_at: datetime
    start_frame: int = Field(description='The frame at which this GameEventType starts')
    end_frame: int | None = Field(
        None, description='The frame at which this GameEventType starts'
    )


class Frame_PossessionEventType(Enum):
    BALL_CARRY = 'BC'
    CHALLENGE = 'CH'  # includes dribbles
    CLEARANCE = 'CL'
    CROSS = 'CR'
    PASS = 'PA'
    REBOUND = 'RE'
    SHOT = 'SH'


class Frame_Possession(BaseModel):
    duration_seconds: float | None = Field(alias='duration', default=None)
    end_time_seconds: float | None = Field(alias='end_time', default=None)
    game_clock: str | None = Field(alias='formatted_game_clock')
    game_event_id: int
    game_id: int
    inserted_at: datetime
    possession_event_type: Frame_PossessionEventType
    start_time_seconds: float = Field(alias='start_time')
    updated_at: datetime
    start_frame: int
    end_frame: int | None = None


class PFF_Frame(BaseModel):
    model_config = ConfigDict(populate_by_name=True, arbitrary_types_allowed=True)

    away_players: list[Frame_Player] | None = Field(
        default_factory=list, alias='awayPlayers'
    )
    away_players_with_kalman: list[Frame_Player] | None = Field(
        default_factory=list, alias='awayPlayersSmoothed'
    )
    ball: list[Ball] | None = Field(default_factory=list, alias='balls')
    ball_with_kalman: Ball | None = Field(alias='ballsSmoothed')
    elapsed_seconds: float = Field(alias='periodElapsedTime')
    event: Frame_Event | None = Field(None, alias='game_event')
    event_id: int | None = Field(None, alias='game_event_id')
    frame_id: int = Field(alias='frameNum')
    game_clock_seconds: float = Field(alias='periodGameClockTime')
    match_id: int | None = Field(None, alias='gameRefId')
    generated_time: datetime | None = Field(alias='generatedTime')
    home_ball: int | None = None
    home_players: list[Frame_Player] | None = Field(
        default_factory=list, alias='homePlayers'
    )
    home_players_with_kalman: list[Frame_Player] | None = Field(
        default_factory=list, alias='homePlayersSmoothed'
    )
    period: int = Field(alias='period')
    possession: Frame_Possession | None = Field(None, alias='possession_event')
    possession_id: int | None = Field(None, alias='possession_event_id')
    sequence: int | None = None
    version: str | None
    video_time_milli: float = Field(alias='videoTimeMs')

    def __repr__(self):
        return f'Frame {self.frame_id} - {self.game_clock_seconds}'



def flatten_dict(dict_obj: dict, parent_key='', sep='.'):
    """Flattens a nested dictionary.

    :param dict_obj: The dictionary to flatten.
    :param parent_key: The base key string (used for recursion).
    :param sep: The separator between keys in the flattened dictionary.
    :return: A flattened dictionary.
    """
    items = []

    for k, v in dict_obj.items():
        new_key = f'{parent_key}{sep}{k}' if parent_key else k
        if isinstance(v, dict):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))

    return dict(items)



from matplotlib.axes import Axes
from mplsoccer.pitch import Pitch

from ..config import PITCH_LENGTH, PITCH_WIDTH
from ..providers.pff.schema.event import PFF_ShootingEvent
from .pitch import get_pitch

shot_plot_options = {
    'ms': 8,
    'alpha': 1,
    'mec': '#000000',
    'mew': 1,
    'linewidth': 2,
    'fillstyle': None,
    'marker': 'o',
}

shot_arrows_options = {
    'width': 0.5,
    'headwidth': 10,
    'headlength': 10,
    'color': 'black',
}


def shot(
    shot_event: PFF_ShootingEvent, *, ax: Axes | None = None, pitch: Pitch | None = None
) -> Axes:
    _pitch, _, _ax = get_pitch()
    if not pitch:
        pitch = _pitch
    if not ax:
        ax = _ax
    pitch.draw(ax=ax)

    if shot_event.shotPointX and shot_event.shotPointY:
        x = shot_event.shotPointX + PITCH_LENGTH / 2
        y = shot_event.shotPointY + PITCH_WIDTH / 2
        pitch.arrows(x, y, PITCH_LENGTH, PITCH_WIDTH / 2, **shot_arrows_options, ax=ax)
        ax.plot(x, y, mfc=None, **shot_plot_options)
    return ax



from .api_client import execute_query, get_client
from .queries import queries


def get_available_competitions(
    *, pff_api_url: str | None = None, pff_api_key: str | None = None
) -> dict:
    query = queries['available_competitions']
    client = get_client(pff_api_url, pff_api_key)
    return execute_query(query, client=client)


def get_competitions(
    *, pff_api_url: str | None = None, pff_api_key: str | None = None
) -> dict:
    query = queries['competitions']
    client = get_client(pff_api_url, pff_api_key)
    return execute_query(query, client=client)


def get_competition(
    competition_id: int,
    *,
    pff_api_url: str | None = None,
    pff_api_key: str | None = None,
) -> dict:
    query = queries['competition']
    variables = {'id': competition_id}
    client = get_client(pff_api_url, pff_api_key)
    return execute_query(query, variables, client=client)



"""Enhances raw PFF event data with data from frames."""

from typing import TypeAlias

from ..providers.pff.schema.event import (
    PFF_BallCarryEvent,
    PFF_ChallengeEvent,
    PFF_ChallengeType,
    PFF_ClearanceEvent,
    PFF_CrossEvent,
    PFF_Event,
    PFF_EventType,
    PFF_Game,
    PFF_PassingEvent,
    PFF_PassOutcomeType,
    PFF_PossessionEvent,
    PFF_Roster,
    PFF_ShootingEvent,
)
from ..providers.pff.schema.tracking import (
    PFF_Frame,
    PFF_Frame_GameEventType,
)

PlayerIdShirt: TypeAlias = dict[str, tuple[int, bool]]


def get_enhanceable_event_ids(
    events: list[PFF_Event],
) -> tuple[list[int], list[int]]:
    """Returns the list of enhanceable events.

    Args:
        events: A list of PFF_Event objects.

    Returns:
        A tuple containing two lists:
            - A list of PFF_Event ids that are enhanceable.
            - A list of PFF_PossessionEvent ids that are enhanceable.
    """
    event_ids = [
        int(event.id)
        for event in events
        if event.gameEventType == PFF_EventType.BALL_OUT_OF_PLAY
    ]

    possession_event_ids = [
        int(pevt.id) for event in events for pevt in event.possessionEvents
    ]

    return event_ids, possession_event_ids


def enhance_match(match: PFF_Game, frames: list[PFF_Frame]) -> PFF_Game:
    """Enhances a PFF match events with frame data."""
    player_id_shirt = _map_player_id_shirt(match.rosters, match.homeTeam.id)
    match.gameEvents = enhance(match.gameEvents, frames, player_id_shirt)
    return match


def enhance(
    events: list[PFF_Event],
    frames: list[PFF_Frame],
    player_id_shirt: PlayerIdShirt,
) -> list[PFF_Event]:
    """Enhances PFF events with frame data."""
    events_copy = [e.model_copy(deep=True) for e in events]
    possession_events = _map_possession_events_by_id(events_copy)
    pevt_frames = _match_possession_events_and_frames(possession_events, frames)

    modified_events = _add_coordinates_to_events(pevt_frames, frames, player_id_shirt)
    updated_events = _update_events_with_modified_possessions(
        events_copy, modified_events
    )
    updated_events = _update_out_events(updated_events, frames)
    return sorted(updated_events, key=lambda e: float(e.startTime))


def _map_player_id_shirt(rosters: list[PFF_Roster], home_team_id: str) -> PlayerIdShirt:
    """Maps player id to shirt number and if the player belongs to the home team."""
    return {
        roster.player.id: (int(roster.shirtNumber), roster.team.id == home_team_id)
        for roster in rosters
    }


def _map_possession_events_by_id(
    events: list[PFF_Event],
) -> dict[str, PFF_PossessionEvent]:
    """Extracts possession events from events."""
    return {pevt.id: pevt for event in events for pevt in event.possessionEvents}


def _match_possession_events_and_frames(
    pevents: dict[str, PFF_PossessionEvent], frames: list[PFF_Frame]
) -> dict[str, tuple[PFF_PossessionEvent, PFF_Frame]]:
    """Matches possession events with frames."""
    pevt_frames = {}

    for frame in frames:
        poss_id = str(frame.possession_id)
        if poss_id and poss_id in pevents:
            pevt_frames[poss_id] = (pevents[poss_id], frame)

    return pevt_frames


def _add_coordinates_to_events(
    pevt_frames: dict[str, tuple[PFF_PossessionEvent, PFF_Frame]],
    frames: list[PFF_Frame],
    player_id_shirt: PlayerIdShirt,
) -> dict[str, PFF_PossessionEvent]:
    """Adds coordinates to events based on frame data."""
    modified_events = {}

    for pevt_id, (event, frame) in pevt_frames.items():
        event_with_coords = _add_coordinates((event, frame), player_id_shirt)
        if event_with_coords and isinstance(event_with_coords, PFF_PossessionEvent):
            event_with_coords = _look_ahead_and_add_coordinates(
                event_with_coords, frames, player_id_shirt
            )
        if event_with_coords:
            modified_events[pevt_id] = event_with_coords

    return modified_events


def _update_out_events(
    events: list[PFF_Event], frames: list[PFF_Frame]
) -> list[PFF_Event]:
    for event in events:
        if event.gameEventType == PFF_EventType.BALL_OUT_OF_PLAY:
            frame = next((f for f in frames if f.event_id == int(event.id)), None)
            if frame and frame.ball:
                event.endPointX = frame.ball[0].x
                event.endPointY = frame.ball[0].y
    return events


def _add_coordinates(
    event_frame: tuple[PFF_PossessionEvent, PFF_Frame],
    player_id_shirt: PlayerIdShirt,
) -> PFF_Event | PFF_PossessionEvent | None:
    """Adds coordinates to an event."""
    event, frame = event_frame

    if isinstance(event, PFF_PossessionEvent):
        if event.fouls:
            _add_coordinates_to_fouls(event.fouls, frame, player_id_shirt)
        if event.shootingEvent:
            _add_coordinates_to_shooting_event(
                event.shootingEvent, frame, player_id_shirt
            )
        if event.crossEvent:
            _add_coordinates_to_cross_event(event.crossEvent, frame, player_id_shirt)
        if event.passingEvent:
            _add_coordinates_to_passing_event(
                event.passingEvent, frame, player_id_shirt
            )
        if event.clearanceEvent:
            _add_coordinates_to_clearance_event(
                event.clearanceEvent, frame, player_id_shirt
            )
        if event.challengeEvent:
            _add_coordinates_to_challenge_event(
                event.challengeEvent, frame, player_id_shirt
            )
        if event.ballCarryEvent:
            _add_coordinates_to_ball_carry_event(
                event.ballCarryEvent, frame, player_id_shirt
            )
        return event

    return None


def _add_coordinates_to_fouls(
    fouls: list, frame: PFF_Frame, player_id_shirt: PlayerIdShirt
):
    """Adds coordinates to foul events."""
    for foul in fouls:
        coords = _get_player_coordinates(foul.culpritPlayer.id, frame, player_id_shirt)
        if coords:
            foul.foulPointX, foul.foulPointY = coords


def _add_coordinates_to_shooting_event(
    evt: PFF_ShootingEvent, frame: PFF_Frame, player_id_shirt: PlayerIdShirt
):
    """Adds coordinates to shooting events."""
    coords = _get_player_coordinates(evt.shooterPlayer.id, frame, player_id_shirt)
    if coords:
        evt.shotPointX, evt.shotPointY = coords

    if evt.blockerPlayer:
        coords = _get_player_coordinates(evt.blockerPlayer.id, frame, player_id_shirt)
        if coords:
            evt.defenderPointX, evt.defenderPointY = coords

    if evt.saverPlayer:
        coords = _get_player_coordinates(evt.saverPlayer.id, frame, player_id_shirt)
        if coords:
            evt.savePointX, evt.savePointY = coords

    if evt.clearerPlayer:
        coords = _get_player_coordinates(evt.clearerPlayer.id, frame, player_id_shirt)
        if coords:
            evt.defenderPointX, evt.defenderPointY = coords


def _add_coordinates_to_cross_event(
    evt: PFF_CrossEvent, frame: PFF_Frame, player_id_shirt: PlayerIdShirt
):
    """Adds coordinates to cross events."""
    coords = _get_player_coordinates(evt.crosserPlayer.id, frame, player_id_shirt)
    if coords:
        evt.crossPointX, evt.crossPointY = coords


def _add_coordinates_to_passing_event(
    evt: PFF_PassingEvent, frame: PFF_Frame, player_id_shirt: PlayerIdShirt
):
    """Adds coordinates to passing events."""
    coords = _get_player_coordinates(evt.passerPlayer.id, frame, player_id_shirt)
    if coords:
        evt.passPointX, evt.passPointY = coords


def _add_coordinates_to_clearance_event(
    evt: PFF_ClearanceEvent, frame: PFF_Frame, player_id_shirt: PlayerIdShirt
):
    """Adds coordinates to clearance events."""
    coords = _get_player_coordinates(evt.clearancePlayer.id, frame, player_id_shirt)
    if coords:
        evt.clearanceStartPointX, evt.clearanceStartPointY = coords


def _add_coordinates_to_challenge_event(
    evt: PFF_ChallengeEvent, frame: PFF_Frame, player_id_shirt: PlayerIdShirt
):
    """Adds coordinates to challenge events."""
    if evt.challengeWinnerPlayer:
        coords = _get_player_coordinates(
            evt.challengeWinnerPlayer.id, frame, player_id_shirt
        )
        if coords:
            evt.challengePointX, evt.challengePointY = coords

    if evt.challengeType == PFF_ChallengeType.DRIBBLE:
        player = evt.ballCarrierPlayer
        coords = _get_player_coordinates(player.id, frame, player_id_shirt)
        if coords:
            evt.dribbleStartPointX, evt.dribbleEndPointY = coords

        if evt.challengerPlayer:
            player = evt.challengerPlayer
            coords = _get_player_coordinates(player.id, frame, player_id_shirt)
            if coords:
                evt.tackleAttemptPointX, evt.tackleAttemptPointY = coords


def _add_coordinates_to_ball_carry_event(
    evt: PFF_BallCarryEvent, frame: PFF_Frame, player_id_shirt: PlayerIdShirt
):
    """Adds coordinates to ball carry events."""
    player = evt.ballCarrierPlayer
    coords = _get_player_coordinates(player.id, frame, player_id_shirt)
    if coords:
        evt.ballCarryStartPointX, evt.ballCarryStartPointY = coords


def _look_ahead_and_add_coordinates(
    possession_event: PFF_PossessionEvent,
    frames: list[PFF_Frame],
    player_id_shirt: PlayerIdShirt,
) -> PFF_PossessionEvent:
    """Looks ahead to add coordinates to events that require future frame data."""
    event_id = int(possession_event.gameEvent.id)
    p_id = int(possession_event.id)
    frames_sorted = sorted(frames, key=lambda f: f.frame_id)

    curr_frame = next((f for f in frames_sorted if f.possession_id == p_id), None)
    next_frame = next(
        (f for f in frames_sorted if f.event_id and f.event_id > event_id),
        None,
    )

    if possession_event.ballCarryEvent and curr_frame:
        _add_ball_carry_end_coordinates(
            possession_event.ballCarryEvent,
            frames_sorted,
            curr_frame,
            player_id_shirt,
        )

    if next_frame and next_frame.event:
        if next_frame.event.game_event_type == PFF_Frame_GameEventType.END_OF_HALF:
            return possession_event

    if next_frame:
        if possession_event.passingEvent:
            _add_passing_event_next_frame_coords(
                possession_event.passingEvent, next_frame, player_id_shirt
            )
        if possession_event.crossEvent:
            _add_cross_event_next_frame_coords(possession_event.crossEvent, next_frame)
        if possession_event.clearanceEvent:
            _add_clearance_event_next_frame_coords(
                possession_event.clearanceEvent, next_frame
            )

    return possession_event


def _get_frame_by_id(frames: list[PFF_Frame], frame_id: int) -> PFF_Frame | None:
    return next((f for f in frames if f.frame_id == frame_id), None)


def _add_ball_carry_end_coordinates(evt, frames, curr_frame, player_id_shirt):
    """Adds end coordinates to ball carry events."""
    start_frame = _get_frame_by_id(frames, curr_frame.event.start_frame)
    end_frame = _get_frame_by_id(frames, curr_frame.event.end_frame)

    if start_frame:
        player = evt.ballCarrierPlayer
        coords = _get_player_coordinates(player.id, start_frame, player_id_shirt)
        if coords:
            evt.ballCarryStartPointX, evt.ballCarryStartPointY = coords
    if end_frame:
        player = evt.ballCarrierPlayer
        coords = _get_player_coordinates(player.id, end_frame, player_id_shirt)
        if coords:
            evt.ballCarryEndPointX, evt.ballCarryEndPointY = coords


def _add_passing_event_next_frame_coords(evt, next_frame, player_id_shirt):
    """Adds coordinates to passing events from the next frame."""
    if evt.receiverPlayer:
        player = evt.receiverPlayer
        coords = _get_player_coordinates(player.id, next_frame, player_id_shirt)
        if coords:
            evt.receiverPointX, evt.receiverPointY = coords
    elif (
        evt.defenderPlayer
        and evt.passOutcomeType == PFF_PassOutcomeType.DEFENSIVE_INTERCEPTION
    ):
        player = evt.defenderPlayer
        coords = _get_player_coordinates(player.id, next_frame, player_id_shirt)
        if coords:
            evt.defenderPointX, evt.defenderPointY = coords
    elif evt.passOutcomeType == PFF_PassOutcomeType.OUT_OF_PLAY:
        if next_frame.ball:
            ball = next_frame.ball[0]
            evt.outOfPlayPointX, evt.outOfPlayPointY = ball.x, ball.y


def _add_cross_event_next_frame_coords(evt, next_frame):
    """Adds end coordinates to cross events from the next frame."""
    if next_frame.ball:
        ball = next_frame.ball[0]
        evt.crossEndPointX, evt.crossEndPointY = ball.x, ball.y


def _add_clearance_event_next_frame_coords(evt, next_frame):
    """Adds end coordinates to clearance events from the next frame."""
    if next_frame.ball:
        ball = next_frame.ball[0]
        evt.clearanceEndPointX, evt.clearanceEndPointY = ball.x, ball.y


def _get_player_coordinates(
    player_id: str, frame: PFF_Frame, player_id_shirt: PlayerIdShirt
) -> tuple[float, float] | None:
    """Gets the coordinates of a player in a frame."""
    player_info = player_id_shirt.get(player_id)

    if not player_info:
        return None

    shirt, is_home = player_info
    players = frame.home_players if is_home else frame.away_players
    player = next((p for p in players if p.shirt == shirt), None)

    if not player:
        return None

    return player.x, player.y


def _update_events_with_modified_possessions(
    events: list[PFF_Event],
    modified_events: dict[str, PFF_PossessionEvent],
) -> list[PFF_Event]:
    """Updates events with modified possession events."""
    for event in events:
        updated_p_events = []
        for p_event in event.possessionEvents:
            updated_p_events.append(modified_events.get(p_event.id, p_event))
        event.possessionEvents = updated_p_events

    return events


def filter_frames_with_actionable_events(raw_frame: dict) -> dict | None:
    """Filter frames that have actionable events."""
    if raw_frame.get('possession_event_id'):
        return raw_frame

    if raw_frame.get('game_event_id'):
        return raw_frame
        # if raw_frame.get('game_event_type') in [
        #     PFF_Frame_GameEventType.OUT_OF_PLAY.value,
        #     PFF_Frame_GameEventType.END_OF_HALF.value,
        # ]:
        #     return raw_frame
        # if raw_frame.get('frameNum') == raw_frame['game_event']['start_frame']:
        #     return raw_frame

    return None



"""Visualization of tracking data."""

from IPython.display import HTML
from matplotlib.animation import FuncAnimation
from matplotlib.axes import Axes
from matplotlib.pyplot import close

from ..config import PITCH_LENGTH, PITCH_WIDTH
from ..providers.pff.schema.tracking import Frame_Player, PFF_Frame
from .pitch import get_pitch

player_pos_plot_options = {'s': 14**2, 'alpha': 0.8, 'linewidths': 1, 'marker': 'o'}
player_shirt_plot_options = {'ha': 'center', 'va': 'center', 'fontsize': 8}
ball_plot_options = {'alpha': 1.0, 's': 5**2, 'linewidths': 1, 'marker': 'o'}


def plot_player(
    x: float,
    y: float,
    shirt: int | None = None,
    *,
    ax: Axes,
    color=('#333333', '#000000'),
    **kwargs,
):
    """Plots a player on the pitch.

    Parameters
    ----------
    x : float
        x-coordinate of the player
    y : float
        y-coordinate of the player
    shirt : int, optional
        Shirt number of the player
    ax : Axes
        Axes object to plot on
    color : tuple
        Tuple of colors for the player and the shirt number

    Returns
    -------
    ax : Axes
        Axes object with the player plotted
    """
    options = {**player_pos_plot_options, **kwargs}
    x = x + PITCH_LENGTH / 2
    y = y + PITCH_WIDTH / 2
    ax.scatter(x, y, facecolor=color[0], edgecolor=color[1], **options)
    if shirt:
        ax.text(x, y, str(shirt), c=color[1], **player_shirt_plot_options)
    return ax


def _plot_players(
    players: list[Frame_Player], color: tuple[str, str], *, ax: Axes
) -> Axes:
    for player in players:
        plot_player(player.x, player.y, color=color, shirt=player.shirt, ax=ax)
    return ax


def _plot_ball(x: float, y: float, *, ax: Axes, color='#000000', **kwargs):
    """Plots the ball on the pitch.

    Parameters
    ----------
    x : float
        x-coordinate of the ball
    y : float
        y-coordinate of the ball
    ax : Axes
        Axes object to plot on
    color : str
        Color of the ball
    """
    options = {**ball_plot_options, **kwargs}
    x = x + PITCH_LENGTH / 2
    y = y + PITCH_WIDTH / 2
    ax.scatter(x, y, facecolor=color, **options)
    return ax


def view_pff_frame(
    frame: PFF_Frame,
    *,
    pitch_length=PITCH_LENGTH,
    pitch_width=PITCH_WIDTH,
    home_colors=('#FFFFFF', '#000000'),
    away_colors=('#000000', '#FFFFFF'),
    ball_color='#000000',
    ax: Axes | None = None,
) -> Axes:
    _, _, ax = get_pitch(pitch_length=pitch_length, pitch_width=pitch_width, ax=ax)

    if frame.home_players:
        _plot_players(frame.home_players, home_colors, ax=ax)
    if frame.away_players:
        _plot_players(frame.away_players, away_colors, ax=ax)

    if frame.ball:
        ball = frame.ball[0]
        _plot_ball(ball.x, ball.y, color=ball_color, ax=ax)

    return ax


def view_frame_sequence(
    frames: list[PFF_Frame],
    *,
    ax: Axes | None = None,
    pitch_length=PITCH_LENGTH,
    pitch_width=PITCH_WIDTH,
    home_colors=('#FFFFFF', '#000000'),
    away_colors=('#000000', '#FFFFFF'),
    ball_color='#000000',
    fps=25,
):
    """Visualization of a sequence of frames."""
    _, fig, ax = get_pitch(pitch_length=pitch_length, pitch_width=pitch_width, ax=ax)

    home_scatter = ax.scatter(
        [],
        [],
        facecolor=home_colors[0],
        edgecolor=home_colors[1],
        **player_pos_plot_options,
    )
    away_scatter = ax.scatter(
        [],
        [],
        facecolor=away_colors[0],
        edgecolor=away_colors[1],
        **player_pos_plot_options,
    )
    ball_scatter = ax.scatter([], [], facecolor=ball_color, **ball_plot_options)
    shirts = [ax.text(0, 0, '', **player_shirt_plot_options) for _ in range(22)]

    offset_x = PITCH_LENGTH / 2
    offset_y = PITCH_WIDTH / 2

    def update(frame_idx):
        nonlocal shirts

        frame = frames[frame_idx]

        home_positions = [[p.x + offset_x, p.y + offset_y] for p in frame.home_players]
        home_scatter.set_offsets(home_positions)

        away_positions = [[p.x + offset_x, p.y + offset_y] for p in frame.away_players]
        away_scatter.set_offsets(away_positions)

        if frame.ball:
            ball_position = [[frame.ball[0].x + offset_x, frame.ball[0].y + offset_y]]
            ball_scatter.set_offsets(ball_position)
            ball_scatter.set_visible(True)
        else:
            ball_scatter.set_visible(False)

        shirt_index = 0
        for pos, player in zip(home_positions, frame.home_players, strict=False):
            x, y = pos
            shirts[shirt_index].set_position((x, y))
            shirts[shirt_index].set_text(str(player.shirt))
            shirts[shirt_index].set_color(home_colors[1])
            shirts[shirt_index].set_visible(True)
            shirt_index += 1

        for pos, player in zip(away_positions, frame.away_players, strict=False):
            x, y = pos
            shirts[shirt_index].set_position((x, y))
            shirts[shirt_index].set_text(str(player.shirt))
            shirts[shirt_index].set_color(away_colors[1])
            shirts[shirt_index].set_visible(True)
            shirt_index += 1

        for i in range(shirt_index, len(shirts)):
            shirts[i].set_visible(False)

        return home_scatter, away_scatter, ball_scatter, *shirts

    anim = FuncAnimation(
        fig,
        update,
        frames=len(frames),
        blit=True,
        interval=1000 / fps,
        repeat_delay=5000,
    )
    html_output = anim.to_jshtml()
    close(fig)

    return HTML(html_output)



"""Modules for exporting data in different formats."""

from typing import Callable, Literal

from pandas import DataFrame

from ..providers.pff.schema.tracking import PFF_Frame
from .dataframe import pff_frames_to_dataframe
from .gif import export_frame_sequence_as_gif
from .png import export_frame_sequence_as_png

ExportFormat = Literal['gif', 'png', 'dataframe']

__all__ = ['export']


_image_exporters: dict[ExportFormat, Callable] = {
    'gif': export_frame_sequence_as_gif,
    'png': export_frame_sequence_as_png,
    'dataframe': pff_frames_to_dataframe,
}


def export(
    frames: list[PFF_Frame] | PFF_Frame,
    *,
    fmt: ExportFormat,
    fps=25,
    filename: str | None = None,
    **kwargs,
) -> str | list[str] | tuple[DataFrame, DataFrame]:
    """Exports a sequence of frames as a GIF or PNG files."""
    if isinstance(frames, PFF_Frame):
        frames = [frames]
    return _image_exporters[fmt](frames, filename=filename, fps=fps, **kwargs)



from typing import Callable, Literal, TypeVar

import pandas as pd

from ..providers.pff.api.response_types import PFFResponse

# TODO: Add Gandula
# TODO: Fix DataType typing
ResponseT = TypeVar('ResponseT', bound=PFFResponse)

Format = Literal['dict', 'json', 'dataframe', 'gandula', 'typed']
DataType = ResponseT | list[ResponseT]
ResponseType = dict | str | pd.DataFrame | DataType


default_view_options = {
    'exclude_none': True,
    'exclude_defaults': True,
    'exclude_unset': True,
}


def process_dict(response: DataType, **options) -> dict | list:
    if isinstance(response, list):
        return [item.model_dump(**options) for item in response]
    return response.model_dump(**options)


def process_json(response: DataType, **options) -> str:
    if isinstance(response, list):
        return str([item.model_dump(**options) for item in response])
    return response.model_dump_json(**options)


def process_dataframe(response: DataType, **options) -> pd.DataFrame:
    if isinstance(response, list):
        return pd.DataFrame([item.model_dump(**options) for item in response])
    return pd.DataFrame(response.model_dump(**options))


def process_typed(response: DataType) -> DataType:
    return response


def process_gandula(response: DataType, **options) -> DataType:
    return response


# TODO: Fix Callable args
formatter_dispatcher: dict[Format, Callable[[DataType], ResponseType]] = {
    'dict': process_dict,
    'json': process_json,
    'dataframe': process_dataframe,
    'gandula': process_gandula,
    'typed': process_typed,
}


def format_response(response: DataType, fmt: Format = 'dict') -> ResponseType:
    """Function to process responses based on the given formatter.

    :param rseponse: The response to be processed.
    :param fmt: The type of formatter to use.
    :return: The result of the processing function corresponding to the formatter.
    """
    if fmt in formatter_dispatcher:
        return formatter_dispatcher[fmt](response, **default_view_options)
    else:
        raise ValueError(f'Unsupported formatter: {fmt}')



from ..providers.pff.api import competitions
from ..providers.pff.api.response_types import AvailableCompetitions
from ..utils.formatter import Format, ResponseType, format_response


def get_competitions(*, api_url: str | None = None, api_key: str | None = None) -> dict:
    """Retrieves all competitions, in full (long).

    :param api_url: The API URL to connect to.
    :param api_key: The API key for authentication.
    :return: A dictionary containing all competitions.
    """
    return competitions.get_competitions(pff_api_url=api_url, pff_api_key=api_key)


def get_available_competitions(
    *, api_url: str | None = None, api_key: str | None = None, fmt: Format = 'dict'
) -> ResponseType:
    """Retrieves only the available competitions for these credentials.

    :param api_url: The API URL to connect to.
    :param api_key: The API key for authentication.
    :return: A dictionary containing available competitions.
    """
    raw = competitions.get_available_competitions(
        pff_api_url=api_url, pff_api_key=api_key
    )
    response = [
        AvailableCompetitions.model_validate(competition)
        for competition in raw['competitions']
    ]
    return format_response(response, fmt)


def get_competition(
    competition_id: int, *, api_url: str | None = None, api_key: str | None = None
) -> dict:
    """Retrieves a specific competition by its ID.

    :param competition_id: The ID of the competition to retrieve.
    :param api_url: The API URL to connect to.
    :param api_key: The API key for authentication.
    :return: A dictionary containing the competition data.
    """
    return competitions.get_competition(
        competition_id, pff_api_url=api_url, pff_api_key=api_key
    )



"""Exports frames and events as pngs."""

from pathlib import Path

from matplotlib.pyplot import close, savefig

from ..providers.pff.schema.tracking import PFF_Frame
from ..visualization.frame import view_pff_frame


def export_frame_sequence_as_png(
    frames: list[PFF_Frame], filename: str, **kwargs
) -> list[str]:
    """Exports a sequence of frames as PNG files.

    Args:
    -----
    frames (list of dict): List of frames, each containing player and ball positions.
    filename : (str): Path to save the PNG files.
    """
    paths = []

    if filename.endswith('.png'):
        filename = filename[:-4]

    for index, frame in enumerate(frames):
        ax = view_pff_frame(frame)
        fig = ax.get_figure()
        filepath = f'{filename}_{index}.png'
        savefig(filepath, bbox_inches='tight', dpi=100)
        paths.append(Path(filepath).resolve().as_uri())
        close(fig)

    return paths



from .api_client import execute_query, get_client
from .queries import queries


def get_match_events(
    match_id: int, *, pff_api_url: str | None = None, pff_api_key: str | None = None
) -> dict:
    query = queries['game_events']
    variables = {'gameId': match_id}
    client = get_client(pff_api_url, pff_api_key)
    return execute_query(query, variables, client=client)


def get_match_event(
    event_id: int, *, pff_api_url: str | None = None, pff_api_key: str | None = None
) -> dict:
    query = queries['game_event']
    variables = {'id': event_id}
    client = get_client(pff_api_url, pff_api_key)
    return execute_query(query, variables, client=client)


def get_match(
    match_id: int, *, pff_api_url: str | None = None, pff_api_key: str | None = None
) -> dict:
    query = queries['game']
    variables = {'id': match_id}
    client = get_client(pff_api_url, pff_api_key)
    return execute_query(query, variables, client=client)



"""Exports sequences of frames as gifs."""

import io
from pathlib import Path

import imageio
from matplotlib.pyplot import close, savefig
from pygifsicle import optimize

from ..providers.pff.schema.tracking import PFF_Frame
from ..visualization.frame import view_pff_frame


def export_frame_sequence_as_gif(frames: list[PFF_Frame], filename: str, fps=25) -> str:
    """
    Exports a sequence of frames as a GIF file.

    Args:
    ----------
    frames (list of dict): List of frames, each containing player and ball positions.
    filename : (str): Filename to save the GIF.
    fps (int): Frames per second for the GIF.
    """
    images = []

    for frame in frames:
        ax = view_pff_frame(frame)

        fig = ax.get_figure()

        buf = io.BytesIO()
        savefig(buf, format='png')

        buf.seek(0)
        images.append(imageio.mimread(buf))

        close(fig)
        buf.close()

    if not filename.endswith('.gif'):
        filename = f'{filename}.gif'

    with imageio.get_writer(filename, fps=fps, mode='I') as writer:
        for img in images:
            writer.append_data(img)

    optimize(filename)  # shrinks the gif size

    return Path(filename).resolve().as_uri()



import os
from typing import Any

from gql import Client, gql
from gql.transport.httpx import HTTPXTransport

PFF_API_URL = 'https://faraday.pff.com/api'
GQL_CLIENT_TIMEOUT = 60


def get_pff_api_url() -> str:
    """Retrieves the PFF API URL from the environment or defaults to the constant.

    :return: The PFF API URL.
    """
    return os.environ.get('PFF_API_URL', PFF_API_URL)


def get_pff_api_key() -> str:
    """Retrieves the PFF API key from the environment.

    :return: The PFF API key.
    :raises ValueError: If the API key is not found.
    """
    pff_api_key = os.environ.get('PFF_API_KEY')

    if not pff_api_key:
        raise ValueError(
            'A PFF API key is required. Please set it in your environment as'
            ' `PFF_API_KEY`. For example, `export PFF_API_KEY=your_api_key`.'
        )

    return pff_api_key


def build_headers(
    pff_api_key: str, additional_headers: dict[str, str] | None = None
) -> dict[str, str]:
    """Builds the header with the pff key for the GraphQL client.

    :param pff_api_key: The PFF API key for authentication.
    :param additional_headers: Additional headers to include in the request.
    :return: A dictionary containing the headers.
    """
    headers = {'x-api-key': pff_api_key}
    if additional_headers:
        headers.update(additional_headers)
    return headers


def get_client(
    pff_api_url: str | None = None,
    pff_api_key: str | None = None,
    *,
    headers: dict[str, str] | None = None,
    timeout: int = GQL_CLIENT_TIMEOUT,
) -> Client:
    """Creates and returns a configured GraphQL client.

    :param pff_api_url: The API URL to connect to. Defaults to the environment variable
        PFF_API_URL.
    :param pff_api_key: The API key for authentication. Defaults to env variable
        PFF_API_KEY.
    :param headers: Additional headers to include in the request.
    :param timeout: Timeout for GraphQL queries in seconds. Defaults to 60 seconds.
    :return: An instance of gql.Client configured with the provided settings.
    :raises ValueError: If the API key is not provided or is not a string.
    """
    pff_api_url = pff_api_url or get_pff_api_url()
    pff_api_key = pff_api_key or get_pff_api_key()

    headers = build_headers(pff_api_key, headers)
    transport = HTTPXTransport(url=pff_api_url, headers=headers, timeout=timeout)

    return Client(
        transport=transport, fetch_schema_from_transport=False, execute_timeout=timeout
    )


def execute_query(
    query: str, variables: dict[str, Any] | None = None, *, client: Client | None = None
) -> dict:
    """Executes a GraphQL query using the provided client.

    :param query: The query to execute.
    :param variables: Variables to include in the query.
    :param client: The GraphQL client to use for the query.
    :return: The result of the query.
    """
    if not client:
        client = get_client()

    try:
        return client.execute(
            gql(query),
            variable_values=variables,
            serialize_variables=True,
            parse_result=True,
        )
    except Exception as exc:
        raise Exception(
            f'There was an error trying to execute the following query:\n'
            f'Query: {query}\n'
            f'Variables: {variables}'
        ) from exc
